# -*- coding: utf-8 -*-
# :Project:   pglast -- Serialization logic
# :Created:   mer 02 ago 2017 15:46:11 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017, 2018, 2019, 2020, 2021, 2022 Lele Gaifax
#

from contextlib import contextmanager
from io import StringIO
from re import compile
from sys import stderr

from . import ast, parse_plpgsql, parse_sql
from .keywords import RESERVED_KEYWORDS, TYPE_FUNC_NAME_KEYWORDS
from .printers import get_printer_for_node, get_special_function


is_simple_name = compile(r'[a-z_][a-z0-9_]*$').match
"Determine whether a name is simple enough to not require being double quoted."


def maybe_double_quote_name(name):
    """Possibly enclose `name` within double quotes.

    When `name` is not entirely composed by lower case letters, digits or underscores, or it
    matches one of the PostgreSQL keywords, it must be emitted between double quotes.
    """

    if ((not is_simple_name(name)
         or name in RESERVED_KEYWORDS
         or name in TYPE_FUNC_NAME_KEYWORDS)):
        name = '"%s"' % name.replace('"', '""')
    return name


class OutputStream(StringIO):
    "A stream that has a concept of a *pending separator* between consecutive writes."

    def __init__(self):
        super().__init__()
        self.pending_separator = False
        self.last_emitted_char = ' '

    def show(self, where=stderr):  # pragma: no cover
        "Emit current stream content, by default to `stderr`, to aid debugging."

        where.write(self.getvalue())
        where.write('\n')

    def separator(self):
        """Possibly insert a single whitespace character before next output.

        When the last character written is not a space, set the `pending_separator` flag to
        ``True``: the next call to :meth:`write` will prepend a single whitespace to its
        argument if that begins with an alphanumeric character.
        """

        if not self.last_emitted_char.isspace():
            self.pending_separator = True

    def maybe_write_space(self, nextc=None, _special_chars=set("""_*+/-"'=<>$@""")):
        """Emit a space when needed.

        :param nextc: either None or the next character that will be written
        :return: the number of characters written to the stream

        If the last character written was neither a space nor an open parentheses, and `nextc`
        is either ``None`` or a *special character*, then emit a single whitespace.
        """

        lec = self.last_emitted_char
        if not lec.isspace() and lec != '(':
            if nextc is None or nextc.isalnum() or nextc in _special_chars:
                return self.write(' ')
        return 0

    def write(self, s):
        """Emit string `s`.

        :param str s: the string to emit
        :return: the number of characters written to the stream

        When `s` is not empty and `pending_separator` is ``True`` and the first character of
        `s` is alphanumeric, emit a single whitespace before writing out `s` and then reset
        `pending_separator` to ``False``.
        """

        count = 0
        if s:
            if self.pending_separator:
                if s != ' ':
                    self.maybe_write_space(s[0])
                self.pending_separator = False
            count = super().write(s)
            self.last_emitted_char = s[-1]

        return count

    def writes(self, s):
        "Shortcut for ``self.write(s); self.separator()``."

        count = self.write(s)
        self.separator()
        return count

    def swrite(self, s):
        "Shortcut for ``self.maybe_write_space(s[0]); self.write(s)``."

        count = self.maybe_write_space(s[0])
        return count + self.write(s)

    def swrites(self, s):
        "Shortcut for ``self.swrite(s); self.separator()``."

        count = self.swrite(s)
        self.separator()
        return count


class RawStream(OutputStream):
    """Basic SQL parse tree writer.

    :param int separate_statements:
           ``1`` by default, tells how many empty lines should separate statements
    :param bool special_functions:
           ``False`` by default, when ``True`` some functions are treated in a special way and
           emitted as equivalent constructs
    :param bool comma_at_eoln:
           ``False`` by default, when ``True`` put the comma right after each item instead of
           at the beginning of the *next* item line
    :param bool semicolon_after_last_statement:
           ``False`` by default, when ``True`` add a semicolon after the last statement,
           otherwise it is emitted only as a separator between multiple statements
    :param comments: optional sequence of tuples with the comments extracted from the statement
    :param bool remove_pg_catalog_from_functions:
           ``False`` by default, when ``True`` remove the pg_catalog schema from functions

    This augments :class:`OutputStream` and implements the basic machinery needed to serialize
    the *parse tree* produced by :func:`~.parser.parse_sql()` back to a textual representation,
    without any adornment.
    """

    def __init__(self, separate_statements=1, special_functions=False,
                 comma_at_eoln=False, semicolon_after_last_statement=False,
                 comments=None, remove_pg_catalog_from_functions=False):
        super().__init__()
        self.current_column = 0
        self.separate_statements = separate_statements
        self.special_functions = special_functions
        self.comma_at_eoln = comma_at_eoln
        self.semicolon_after_last_statement = semicolon_after_last_statement
        self.comments = comments
        self.remove_pg_catalog_from_functions = remove_pg_catalog_from_functions

    def show(self, where=stderr):  # pragma: no cover
        """Emit also current expression_level and a "pointer" showing current_column."""

        super().show(where)
        if self.current_column:
            where.write('-' * (self.current_column-1))
        where.write('^\n')

    def __call__(self, sql, plpgsql=False):
        """Main entry point: execute :meth:`print_node` on each statement in `sql`.

        :param sql: the SQL statement
        :param bool plpgsql: whether `sql` is really a ``plpgsql`` statement
        :returns: a string with the equivalent SQL obtained by serializing the syntax tree

        The `sql` statement may be either a ``str`` containing the ``SQL`` in textual form, or
        a concrete :class:`.ast.Node` instance or a tuple of them.
        """

        from .visitors import Visitor

        if isinstance(sql, str):
            sql = parse_plpgsql(sql) if plpgsql else parse_sql(sql)
        elif isinstance(sql, ast.Node):
            sql = (sql,)
        elif not (isinstance(sql, tuple) and sql and isinstance(sql[0], ast.Node)):
            raise ValueError("Unexpected value for 'sql', must be either a string,"
                             " an ast.Node or tuple of them, got %r" % type(sql))

        class UpdateAncestors(Visitor):
            def visit(self, ancestors, node):
                node.ancestors = ancestors

        UpdateAncestors()(sql)

        first = True
        for statement in sql:
            if first:
                first = False
            else:
                self.write(';')
                self.newline()
                for _ in range(self.separate_statements):
                    self.newline()
            self.print_node(statement)

        if self.semicolon_after_last_statement:
            self.write(';')

        if self.comments:
            while self.comments:
                self.print_comment(self.comments.pop(0))

        return self.getvalue()

    def dedent(self):
        "Do nothing, shall be overridden by the prettifier subclass."

    def get_printer_for_function(self, name):
        """Look for a specific printer for function `name` in :data:`SPECIAL_FUNCTIONS`.

        :param str name: the qualified name of the function
        :returns: either ``None`` or a callable

        When the option `special_functions` is ``True``, return the printer function associated
        with `name`, if present. In all other cases return ``None``.
        """

        if self.special_functions:
            return get_special_function(name)

    def indent(self, amount=0, relative=True):
        "Do nothing, shall be overridden by the prettifier subclass."

    def newline(self):
        "Emit a single whitespace, shall be overridden by the prettifier subclass."

        self.separator()

    def space(self, count=1):
        "Emit a single whitespace, shall be overridden by the prettifier subclass."

        self.separator()

    @contextmanager
    def push_indent(self, amount=0, relative=True):
        "Create a no-op context manager, shall be overridden by the prettifier subclass."

        yield

    @contextmanager

    def expression(self, need_parens):
        """Create a context manager usable to conditionally wrap something within parentheses.

        :param bool need_parens: whether to emit opening and closing parentheses

        This method used to be used for *subexpressions*, but now is used to emit almost
        anything that goes within parentheses.
        """

        if need_parens:
            self.write('(')
        yield
        if need_parens:
            self.write(')')

    def _concat_nodes(self, nodes, sep=' ', are_names=False):

        """Concatenate given `nodes`, using `sep` as the separator.

        :param scalars: a sequence of nodes
        :param str sep: the separator between them
        :param bool are_names:
               whether the nodes are actually *names*, which possibly require to be enclosed
               between double-quotes
        :returns: a string

        Use a temporary :class:`RawStream` instance to print the list of nodes and return the
        result.
        """

        rawstream = RawStream(
            special_functions=self.special_functions,
            remove_pg_catalog_from_functions=self.remove_pg_catalog_from_functions)
        rawstream.print_list(nodes, sep, are_names=are_names, standalone_items=False)
        return rawstream.getvalue()

    def write_quoted_string(self, s):
        "Emit the string `s` as a single-quoted literal constant."

        self.write("'%s'" % s.replace("'", "''"))

    def _print_scalar(self, value, is_name, is_symbol):
        "Print the scalar `value`, special-casing string literals."

        is_string = isinstance(value, str)
        if is_symbol and is_string:
            self.write(value)
        elif is_name and is_string:
            self.write(maybe_double_quote_name(value))
        elif is_string:
            self.write_quoted_string(value)
        else:
            self.write(str(value))

    def print_comment(self, comment):
        "Print the given `comment`, unconditionally in the ``C`` syntax, joining all lines."

        is_sql_comment = comment.text.startswith('--')
        if is_sql_comment:
            text = comment.text[2:].strip()
        else:
            text = comment.text[2:-2].strip()
        if text:
            self.swrite('/*')
            lines = text.splitlines()
            if len(lines) > 1:
                with self.push_indent():
                    for line in lines:
                        self.write(line)
                        self.write(' ')
            else:
                self.write(lines[0])
            self.writes('*/')

    def print_name(self, nodes, sep='.'):
        "Helper method, execute :meth:`print_node` or :meth:`print_list` as needed."

        if isinstance(nodes, (list, tuple)):
            self.print_list(nodes, sep, standalone_items=False, are_names=True)
        elif isinstance(nodes, ast.Node):
            self.print_node(nodes, is_name=True)
        else:
            self._print_scalar(nodes, is_name=True, is_symbol=False)
        self.separator()

    def print_symbol(self, nodes, sep='.'):
        "Helper method, execute :meth:`print_node` or :meth:`print_list` as needed."

        if isinstance(nodes, (list, tuple)):
            self.print_list(nodes, sep, standalone_items=False, are_names=True, is_symbol=True)
        elif isinstance(nodes, ast.Node):
            self.print_node(nodes, is_name=True, is_symbol=True)
        else:
            self._print_scalar(str(nodes), is_name=True, is_symbol=True)
        self.separator()

    def print_node(self, node, is_name=False, is_symbol=False):
        """Lookup the specific printer for the given `node` and execute it.

        :param node: an instance of :class:`~.ast.Node`
        :param bool is_name:
               whether this is a *name* of something, that may need to be double quoted
        :param bool is_symbol:
               whether this is the name of an *operator*, should not be double quoted
        """

        if self.comments:
            if hasattr(node, 'location'):
                node_location = getattr(node, 'location')
            elif hasattr(node, 'stmt_location'):
                node_location = getattr(node, 'stmt_location')
            else:
                node_location = None
            if node_location is not None:
                nextc = self.comments[0]
                if nextc.location <= node_location:
                    self.print_comment(self.comments.pop(0))
                    while self.comments and self.comments[0].continue_previous:
                        self.print_comment(self.comments.pop(0))

        if is_name and isinstance(node, (list, tuple)):
            self.print_list(node, '.', standalone_items=False, are_names=True)
        elif isinstance(node, ast.Node):
            printer = get_printer_for_node(node)
            if is_name and isinstance(node, ast.String):
                printer(node, self, is_name=is_name, is_symbol=is_symbol)
            else:
                printer(node, self)
        else:
            self._print_scalar(node, is_name, is_symbol)

        self.separator()

    def _is_pg_catalog_func(self, items):
        return (
            self.remove_pg_catalog_from_functions
            and len(items) > 1
            and isinstance(items, (list, tuple))
            and items[0].ancestors.parent.member == 'funcname'
            and items[0].sval == 'pg_catalog'
            # The list contains all functions that cannot be found without an
            # explicit pg_catalog schema. ie:
            # position(a,b) is invalid but pg_catalog.position(a,b) is fine
            and items[1].sval not in ('position', 'xmlexists')
        )

    def _print_items(self, items, sep, newline, are_names=False, is_symbol=False,
                     item_needs_parens=None):
        first = 1 if self._is_pg_catalog_func(items) else 0
        last = len(items) - 1
        for idx, item in enumerate(items):
            if idx < first:
                continue
            if idx > first:
                if sep == ',' and self.comma_at_eoln:
                    self.write(sep)
                    if newline:
                        self.newline()
                    else:
                        self.write(' ')
                else:
                    if not are_names:
                        if newline:
                            self.newline()
                    if sep:
                        self.write(sep)
                        if sep != '.':
                            self.write(' ')
            if item is None:
                self.write('None')
            else:
                if item_needs_parens is None:
                    needs_parens = False
                else:
                    needs_parens = item_needs_parens(item)
                with self.expression(needs_parens):
                    self.print_node(item, is_name=are_names,
                                    is_symbol=is_symbol and idx == last)

    def print_list(self, nodes, sep=',', relative_indent=None, standalone_items=None,
                   are_names=False, is_symbol=False, item_needs_parens=None):
        """Execute :meth:`print_node` on all the `nodes`, separating them with `sep`.

        :param nodes: a sequence of :class:`~.ast.Node` instances
        :param str sep: the separator between them
        :param bool relative_indent:
               if given, the relative amount of indentation to apply before the first item, by
               default computed automatically from the length of the separator `sep`
        :param bool standalone_items: whether a newline will be emitted before each item
        :param bool are_names:
               whether the nodes are actually *names*, which possibly require to be enclosed
               between double-quotes
        :param bool is_symbol:
               whether the nodes are actually a *symbol* such as an *operator name*, in which
               case the last one must be printed verbatim (e.g. ``"MySchema".===``)
        :param item_needs_parens: either ``None`` or a callable
        """

        if relative_indent is None:
            if are_names or is_symbol:
                relative_indent = 0
            else:
                relative_indent = (-(len(sep) + 1)
                                   if sep and (sep != ',' or not self.comma_at_eoln)
                                   else 0)

        if standalone_items is None:
            standalone_items = not all(isinstance(n, (ast.A_Const, ast.ColumnRef,
                                                      ast.SetToDefault, ast.RangeVar))
                                       for n in nodes)

        with self.push_indent(relative_indent):
            self._print_items(nodes, sep, standalone_items, are_names=are_names,
                              is_symbol=is_symbol, item_needs_parens=item_needs_parens)

    def print_lists(self, lists, sep=',', relative_indent=None, standalone_items=None,
                    are_names=False, sublist_open='(', sublist_close=')', sublist_sep=',',
                    sublist_relative_indent=None):
        """Execute :meth:`print_list` on all the `lists` items.

        :param lists: a sequence of sequences of :class:`~.ast.Node` instances
        :param str sep: passed as is to :meth:`print_list`
        :param bool relative_indent: passed as is to :meth:`print_list`
        :param bool standalone_items: passed as is to :meth:`print_list`
        :param bool are_names: passed as is to :meth:`print_list`
        :param str sublist_open: the string that will be emitted before each sublist
        :param str sublist_close: the string that will be emitted after each sublist
        :param str sublist_sep: the separator between them each sublist
        :param bool sublist_relative_indent:
               if given, the relative amount of indentation to apply before the first sublist,
               by default computed automatically from the length of the separator `sublist_sep`
        """

        if sublist_relative_indent is None:
            sublist_relative_indent = (-(len(sublist_sep) + 1)
                                       if sublist_sep and (sublist_sep != ','
                                                           or not self.comma_at_eoln)
                                       else 0)

        with self.push_indent(sublist_relative_indent):
            self.write(sublist_open)
            first = True
            for lst in lists:
                if first:
                    first = False
                else:
                    if self.comma_at_eoln:
                        self.write(sublist_sep)
                        self.newline()
                        self.write(sublist_open)
                    else:
                        self.newline()
                        self.write(sublist_sep)
                        self.write(' ')
                        self.write(sublist_open)
                self.print_list(lst, sep, relative_indent, standalone_items, are_names)
                self.write(sublist_close)


class IndentedStream(RawStream):
    r"""Indented SQL parse tree writer.

    :param int compact_lists_margin:
           an integer value that, if given, is used to print lists on a single line, when they
           do not exceed the given margin on the right
    :param int split_string_literals_threshold:
           an integer value that, if given, is used as the threshold beyond that a string
           literal gets splitted in successive chunks of that length
    :param \*\*options: other options accepted by :class:`RawStream`

    This augments :class:`RawStream` to emit a prettified representation of a *parse tree*.
    """

    def __init__(self, compact_lists_margin=None, split_string_literals_threshold=None,
                 **options):
        super().__init__(**options)
        self.compact_lists_margin = compact_lists_margin
        self.split_string_literals_threshold = split_string_literals_threshold
        self.current_indent = 0
        self.indentation_stack = []

    def show(self, where=stderr):  # pragma: no cover
        "Emit also current_indent and indentation_stack."

        where.write('current_indent=%d\n' % self.current_indent)
        where.write('indentation_stack=%r\n' % self.indentation_stack)
        super().show(where)

    def dedent(self):
        "Pop the indentation level from the stack and set `current_indent` to that."

        self.current_indent = self.indentation_stack.pop()

    def indent(self, amount=0, relative=True):
        """Push current indentation level to the stack, then set it adding `amount` to the
        `current_column` if `relative` is ``True`` otherwise to `current_indent`.
        """

        self.indentation_stack.append(self.current_indent)
        base_indent = (self.current_column if relative else self.current_indent)
        assert base_indent + amount >= 0, f'base_indent={base_indent} amount={amount}'
        self.current_indent = base_indent + amount

    @contextmanager
    def push_indent(self, amount=0, relative=True):
        """Create a context manager that calls :meth:`indent` and :meth:`dedent` around a block
        of code.

        This is just an helper to simplify code that adjusts the indentation level:

        .. code-block:: python

          with output.push_indent(4):
              # code that emits something with the new indentation
        """

        if self.pending_separator and relative:
            amount += 1
        if self.current_column == 0 and relative:
            amount += self.current_indent
        self.indent(amount, relative)
        yield
        self.dedent()

    def newline(self):
        "Emit a newline."

        self.write('\n')

    def space(self, count=1):
        "Emit consecutive spaces."

        self.write(' '*count)

    def print_comment(self, comment):
        is_sql_comment = comment.text.startswith('--')
        if is_sql_comment:
            text = comment.text[2:].strip()
        else:
            text = comment.text[2:-2].strip()
        if text:
            ci = self.current_indent
            cc = self.current_column
            if self.tell() != 0 and comment.at_start_of_line and not comment.continue_previous:
                self.newline()
            lines = comment.text.splitlines()
            if len(lines) > 1:
                with self.push_indent():
                    for line in lines:
                        self.write(line)
                        self.newline()
            else:
                self.write(lines[0])
                if is_sql_comment:
                    self.newline()
                else:
                    self.space()
            self.current_column = cc
            self.current_indent = ci

    def print_list(self, nodes, sep=',', relative_indent=None, standalone_items=None,
                   are_names=False, is_symbol=False, item_needs_parens=None):
        """Execute :meth:`print_node` on all the `nodes`, separating them with `sep`.

        :param nodes: a sequence of :class:`~.ast.Node` instances
        :param str sep: the separator between them
        :param bool relative_indent:
               if given, the relative amount of indentation to apply before the first item, by
               default computed automatically from the length of the separator `sep`
        :param bool standalone_items: whether a newline will be emitted before each item
        :param bool are_names:
               whether the nodes are actually *names*, which possibly require to be enclosed
               between double-quotes
        :param bool is_symbol:
               whether the nodes are actually an *operator name*, in which case the last one
               must be printed verbatim (such as ``"MySchema".===``)
        """

        if standalone_items is None:
            clm = self.compact_lists_margin
            if clm is not None and clm > 0:
                rawlist = self._concat_nodes(nodes, sep, are_names)
                if self.current_column + len(rawlist) < clm:
                    self.write(rawlist)
                    return

            standalone_items = not all(isinstance(n, (ast.A_Const, ast.ColumnRef,
                                                      ast.SetToDefault, ast.RangeVar))
                                       for n in nodes)

        if (((sep != ',' or not self.comma_at_eoln)
             and len(nodes) > 1
             and len(sep) > 1
             and relative_indent is None
             and not are_names
             and not is_symbol
             and standalone_items)):
            self.write(' '*(len(sep) + 1))  # separator added automatically

        super().print_list(nodes, sep, relative_indent, standalone_items, are_names, is_symbol,
                           item_needs_parens)

    def write_quoted_string(self, s):
        """Emit the string `s` possibly splitted in successive chunks.

        When the ``split_string_literals_threshold`` option is greater than 0 and the length of
        `s` exceeds that value, split the string into multiple chunks.
        """

        sslt = self.split_string_literals_threshold
        if sslt is None or sslt <= 0:
            super().write_quoted_string(s)
        else:
            multiline = '\n' in s
            if multiline:
                self.write('E')
            with self.push_indent():
                while True:
                    chunk = s[:sslt]
                    s = s[sslt:]
                    # Avoid splitting on backslash
                    while chunk.endswith("\\"):
                        chunk += s[0]
                        s = s[1:]
                    chunk = chunk.replace("'", "''")
                    if multiline:
                        chunk = chunk.replace("\\", "\\\\")
                        chunk = chunk.replace("\n", "\\n")
                    self.write("'%s'" % chunk)
                    if s:
                        self.newline()
                    else:  # pragma: no cover
                        break

    def write(self, s):
        """Write string `s` to the stream, adjusting the `current_column` accordingly.

        :param str s: the string to emit
        :return: the number of characters written to the stream

        If `s` is a newline character (``\\n``) set `current_column` to 0. Otherwise when
        `current_column` is 0 and `current_indent` is greater than 0 emit a number of
        whitespaces *before* emitting `s`, to indent it as expected.
        """

        if s and s != '\n' and self.current_column == 0 and self.current_indent > 0:
            self.current_column = super().write(' ' * self.current_indent)

        count = super().write(s)
        if s == '\n':
            self.current_column = 0
        else:
            self.current_column += count

        return count



# -*- coding: utf-8 -*-
# :Project:   pglast -- Special functions
# :Created:   mer 22 nov 2017 08:34:34 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017, 2018, 2021, 2022 Lele Gaifax
#

from . import special_function

# reminder for not yet implemented special functions:
#
# treat(1 as int)
# 'foo' is nfc normalized
# 'foo' is not nfc normalized


def _print_trim(where, node, output):
    output.write('trim({}'.format(where))
    if len(node.args) > 1:
        output.write(' ')
        output.print_node(node.args[1])
    output.write(' FROM ')
    output.print_node(node.args[0])
    output.write(')')


@special_function('pg_catalog.btrim')
def btrim(node, output):
    """
    Emit function ``pg_catalog.btrim('  abc  ')`` as ``trim(BOTH FROM '  abc  ')``
    and ``pg_catalog.btrim('xxabcxx', 'x')`` as ``trim(BOTH 'x' FROM 'xxabcxx')``.
    """
    _print_trim('BOTH', node, output)


@special_function('pg_catalog.pg_collation_for')
def pg_collation_for(node, output):
    "Emit function ``pg_catalog.pg_collation_for(x)`` as ``COLLATION FOR (x)``."
    output.write('COLLATION FOR ')
    with output.expression(True):
        output.print_node(node.args[0])


@special_function('pg_catalog.extract')
def extract(node, output):
    """
    Emit function ``pg_catalog.extract(field, timestamp)`` as
    ``EXTRACT(field FROM timestamp).``.
    """
    output.write('EXTRACT')
    with output.expression(True):
        output.write(node.args[0].val.sval.upper())
        output.write(' FROM ')
        output.print_node(node.args[1])


@special_function('pg_catalog.ltrim')
def ltrim(node, output):
    """
    Emit function ``pg_catalog.ltrim('  abc  ')`` as ``trim(LEADING FROM '  abc  ')``
    and ``pg_catalog.ltrim('xxabcxx', 'x')`` as ``trim(LEADING 'x' FROM 'xxabcxx').``
    """
    _print_trim('LEADING', node, output)


# normalize(U&'\0061\0308bc', NFC)
@special_function('pg_catalog.normalize')
def normalize(node, output):
    """
    Emit function ``pg_catalog.normalize(a)`` as ``normalize(x)`` and function
    ``pg_catalog.normalize('a','b')`` as ``normalize('a', b)``.
    """
    output.write('normalize')
    with output.expression(True):
        output.print_node(node.args[0])
        if len(node.args) > 1:
            output.write(', ')
            output.write(node.args[1].val.sval.upper())


@special_function('pg_catalog.overlaps')
def overlaps(node, output):
    "Emit function ``pg_catalog.overlaps(a, b, c, d)`` as ``(a, b) OVERLAPS (c, d)``."
    with output.expression(True):
        output.print_list((node.args[0], node.args[1]), standalone_items=False)
    output.write(' OVERLAPS ')
    with output.expression(True):
        output.print_list((node.args[2], node.args[3]), standalone_items=False)


@special_function('pg_catalog.overlay')
def overlay(node, output):
    """
    Emit function ``pg_catalog.overlay('Txxxxas','hom', 2, 4)`` as
    ``overlay('Txxxxas' PLACING 'hom' FROM 2 FOR 4)``."
    """
    output.write('overlay')
    with output.expression(True):
        output.print_node(node.args[0])
        output.write(' PLACING ')
        output.print_node(node.args[1])
        output.write(' FROM ')
        output.print_node(node.args[2])
        output.write(' FOR ')
        output.print_node(node.args[3])


@special_function('pg_catalog.position')
def position(node, output):
    "Emit function ``pg_catalog.position('abcd', 'a')`` as ``position('a' IN 'abcd')``."
    output.write('position')
    with output.expression(True):
        output.print_node(node.args[1])
        output.write(' IN ')
        output.print_node(node.args[0])


@special_function('pg_catalog.rtrim')
def rtrim(node, output):
    """
    Emit function ``pg_catalog.rtrim('  abc  ')`` as ``trim(TRAILING FROM '  abc  ')``
    and ``pg_catalog.rtrim('xxabcxx', 'x')`` as ``trim(TRAILING 'x' FROM 'xxabcxx')``
    """
    _print_trim('TRAILING', node, output)


@special_function('pg_catalog.substring')
def substring(node, output):
    """
    Emit function ``pg_catalog.substring('Txxxxas', 2, 4)`` as
    ``substring('Txxxxas' FROM 2 FOR 4)`` and ``pg_catalog.substring('blabla', 2)``
    as ``substring('blabla' FROM 2)``.
    """
    output.write('substring')
    with output.expression(True):
        output.print_node(node.args[0])
        output.write(' FROM ')
        output.print_node(node.args[1])
        if len(node.args) > 2:
            output.write(' FOR ')
            output.print_node(node.args[2])


@special_function('pg_catalog.timezone')
def timezone(node, output):
    """
    Emit function ``pg_catalog.timezone(tz, timestamp)`` as ``timestamp AT TIME ZONE tz``.
    """
    output.print_node(node.args[1])
    output.write(' AT TIME ZONE ')
    output.print_node(node.args[0])


@special_function('pg_catalog.xmlexists')
def xmlexists(node, output):
    "Emit function ``pg_catalog.xmlexists(x, y)`` as ``xmlexists(x PASSING BY REF y)``."
    output.write('xmlexists')
    with output.expression(True):
        output.print_node(node.args[0])
        output.write(' PASSING BY REF ')
        output.print_node(node.args[1])



# -*- coding: utf-8 -*-
# :Project:   pglast -- PostgreSQL Languages AST
# :Created:   mer 02 ago 2017 15:11:02 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017, 2018, 2019, 2021, 2022 Lele Gaifax
#

from collections import namedtuple

from . import enums
from .error import Error
try:
    from .parser import fingerprint, get_postgresql_version, parse_sql, scan, split
except ModuleNotFoundError:  # pragma: no cover
    # bootstrap
    pass


# This is injected automatically at release time
__version__ = 'v5.0.dev0'
"Package's version."

__author__ = 'Lele Gaifax <lele@metapensiero.it>'
"Package's author."


def parse_plpgsql(statement):
    """Parse the given ``PLPGSQL`` `statement` and return its abstract syntax tree.

    Note that this currently returns the raw tree, represented by plain Python structures such
    as lists, dictionaries and scalar values.
    """

    from json import loads
    from .parser import parse_plpgsql_json

    return loads(parse_plpgsql_json(statement))


Comment = namedtuple('Comment', ('location', 'text', 'at_start_of_line', 'continue_previous'))
"A structure to carry information about a single SQL comment."


def _extract_comments(statement):
    lines = []
    lofs = 0
    for line in statement.splitlines(True):
        llen = len(line)
        lines.append((lofs, lofs+llen, line))
        lofs += llen
    comments = []
    continue_previous = False
    for token in scan(statement):
        if token.name in ('C_COMMENT', 'SQL_COMMENT'):
            for bol_ofs, eol_ofs, line in lines:
                if bol_ofs <= token.start < eol_ofs:
                    break
            else:  # pragma: no cover
                raise RuntimeError('Uhm, logic error!')
            at_start_of_line = not line[:token.start - bol_ofs].strip()
            text = statement[token.start:token.end+1]
            comments.append(Comment(token.start, text, at_start_of_line, continue_previous))
            continue_previous = True
        else:
            continue_previous = False
    return comments


def prettify(statement, safety_belt=False, preserve_comments=False, **options):
    r"""Render given `statement` into a prettified format.

    :param str statement: the SQL statement(s)
    :param bool safety_belt: whether to perform a safe check against bugs in pglast's
                             serialization
    :param bool preserve_comments: whether comments shall be preserved, defaults to not
    :param \*\*options: any keyword option accepted by :class:`~.stream.IndentedStream`
                        constructor
    :returns: a string with the equivalent prettified statement(s)

    When `safety_belt` is ``True``, the resulting statement is parsed again and its *AST*
    compared with the original statement: if they don't match, a warning is emitted and the
    original statement is returned. By default it is ``False``, so no double check is done.
    """

    # Intentional lazy imports, so the modules are loaded on demand

    from .stream import IndentedStream
    from . import printers  # noqa

    if preserve_comments:
        options['comments'] = _extract_comments(statement)

    orig_pt = parse_sql(statement)
    prettified = IndentedStream(**options)(orig_pt)
    if safety_belt:
        from logging import getLogger
        import warnings

        try:
            pretty_pt = parse_sql(prettified)
        except Error as e:
            logger = getLogger(__file__)
            logger.warning("Detected a bug in pglast serialization, original statement:\n\n"
                           "%s\n\nhas been serialized to the following invalid one:\n\n%s",
                           statement, prettified)
            warnings.warn(f"Detected a bug in pglast serialization, please report: {e}",
                          RuntimeWarning)
            return statement

        if pretty_pt != orig_pt:
            logger = getLogger(__file__)
            logger.warning("Detected a non-cosmetic difference between this original"
                           " statement:\n\n%s\n\nand the prettified one:\n\n%s",
                           statement, prettified)
            warnings.warn("Detected a non-cosmetic difference between original and"
                          " prettified statements, please report",  RuntimeWarning)
            return statement

    return prettified


__all__ = ('Error', 'enums', 'fingerprint', 'get_postgresql_version',
           'parse_plpgsql', 'parse_sql', 'prettify', 'split')



# -*- coding: utf-8 -*-
# :Project:   pglast -- DO NOT EDIT: automatically extracted from nodes.h @ 15-4.0.0-0-g1ada550
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017-2022 Lele Gaifax
#

from enum import Enum, IntEnum, IntFlag, auto

try:
    from enum import StrEnum
except ImportError:  # pragma: no cover
    # Python < 3.10
    class StrEnum(str, Enum):
        pass


class AggSplit(IntEnum):
    AGGSPLIT_SIMPLE = 0
    AGGSPLIT_INITIAL_SERIAL = 0x02 | 0x04
    AGGSPLIT_FINAL_DESERIAL = 0x01 | 0x08

class AggStrategy(IntEnum):
    AGG_PLAIN = 0
    AGG_SORTED = auto()
    AGG_HASHED = auto()
    AGG_MIXED = auto()

class CmdType(IntEnum):
    CMD_UNKNOWN = 0
    CMD_SELECT = auto()
    CMD_UPDATE = auto()
    CMD_INSERT = auto()
    CMD_DELETE = auto()
    CMD_MERGE = auto()
    CMD_UTILITY = auto()
    CMD_NOTHING = auto()

class JoinType(IntEnum):
    JOIN_INNER = 0
    JOIN_LEFT = auto()
    JOIN_FULL = auto()
    JOIN_RIGHT = auto()
    JOIN_SEMI = auto()
    JOIN_ANTI = auto()
    JOIN_UNIQUE_OUTER = auto()
    JOIN_UNIQUE_INNER = auto()

class LimitOption(IntEnum):
    LIMIT_OPTION_DEFAULT = 0
    LIMIT_OPTION_COUNT = auto()
    LIMIT_OPTION_WITH_TIES = auto()

class NodeTag(IntEnum):
    T_Invalid = 0
    T_IndexInfo = auto()
    T_ExprContext = auto()
    T_ProjectionInfo = auto()
    T_JunkFilter = auto()
    T_OnConflictSetState = auto()
    T_MergeActionState = auto()
    T_ResultRelInfo = auto()
    T_EState = auto()
    T_TupleTableSlot = auto()
    T_Plan = auto()
    T_Result = auto()
    T_ProjectSet = auto()
    T_ModifyTable = auto()
    T_Append = auto()
    T_MergeAppend = auto()
    T_RecursiveUnion = auto()
    T_BitmapAnd = auto()
    T_BitmapOr = auto()
    T_Scan = auto()
    T_SeqScan = auto()
    T_SampleScan = auto()
    T_IndexScan = auto()
    T_IndexOnlyScan = auto()
    T_BitmapIndexScan = auto()
    T_BitmapHeapScan = auto()
    T_TidScan = auto()
    T_TidRangeScan = auto()
    T_SubqueryScan = auto()
    T_FunctionScan = auto()
    T_ValuesScan = auto()
    T_TableFuncScan = auto()
    T_CteScan = auto()
    T_NamedTuplestoreScan = auto()
    T_WorkTableScan = auto()
    T_ForeignScan = auto()
    T_CustomScan = auto()
    T_Join = auto()
    T_NestLoop = auto()
    T_MergeJoin = auto()
    T_HashJoin = auto()
    T_Material = auto()
    T_Memoize = auto()
    T_Sort = auto()
    T_IncrementalSort = auto()
    T_Group = auto()
    T_Agg = auto()
    T_WindowAgg = auto()
    T_Unique = auto()
    T_Gather = auto()
    T_GatherMerge = auto()
    T_Hash = auto()
    T_SetOp = auto()
    T_LockRows = auto()
    T_Limit = auto()
    T_NestLoopParam = auto()
    T_PlanRowMark = auto()
    T_PartitionPruneInfo = auto()
    T_PartitionedRelPruneInfo = auto()
    T_PartitionPruneStepOp = auto()
    T_PartitionPruneStepCombine = auto()
    T_PlanInvalItem = auto()
    T_PlanState = auto()
    T_ResultState = auto()
    T_ProjectSetState = auto()
    T_ModifyTableState = auto()
    T_AppendState = auto()
    T_MergeAppendState = auto()
    T_RecursiveUnionState = auto()
    T_BitmapAndState = auto()
    T_BitmapOrState = auto()
    T_ScanState = auto()
    T_SeqScanState = auto()
    T_SampleScanState = auto()
    T_IndexScanState = auto()
    T_IndexOnlyScanState = auto()
    T_BitmapIndexScanState = auto()
    T_BitmapHeapScanState = auto()
    T_TidScanState = auto()
    T_TidRangeScanState = auto()
    T_SubqueryScanState = auto()
    T_FunctionScanState = auto()
    T_TableFuncScanState = auto()
    T_ValuesScanState = auto()
    T_CteScanState = auto()
    T_NamedTuplestoreScanState = auto()
    T_WorkTableScanState = auto()
    T_ForeignScanState = auto()
    T_CustomScanState = auto()
    T_JoinState = auto()
    T_NestLoopState = auto()
    T_MergeJoinState = auto()
    T_HashJoinState = auto()
    T_MaterialState = auto()
    T_MemoizeState = auto()
    T_SortState = auto()
    T_IncrementalSortState = auto()
    T_GroupState = auto()
    T_AggState = auto()
    T_WindowAggState = auto()
    T_UniqueState = auto()
    T_GatherState = auto()
    T_GatherMergeState = auto()
    T_HashState = auto()
    T_SetOpState = auto()
    T_LockRowsState = auto()
    T_LimitState = auto()
    T_Alias = auto()
    T_RangeVar = auto()
    T_TableFunc = auto()
    T_Var = auto()
    T_Const = auto()
    T_Param = auto()
    T_Aggref = auto()
    T_GroupingFunc = auto()
    T_WindowFunc = auto()
    T_SubscriptingRef = auto()
    T_FuncExpr = auto()
    T_NamedArgExpr = auto()
    T_OpExpr = auto()
    T_DistinctExpr = auto()
    T_NullIfExpr = auto()
    T_ScalarArrayOpExpr = auto()
    T_BoolExpr = auto()
    T_SubLink = auto()
    T_SubPlan = auto()
    T_AlternativeSubPlan = auto()
    T_FieldSelect = auto()
    T_FieldStore = auto()
    T_RelabelType = auto()
    T_CoerceViaIO = auto()
    T_ArrayCoerceExpr = auto()
    T_ConvertRowtypeExpr = auto()
    T_CollateExpr = auto()
    T_CaseExpr = auto()
    T_CaseWhen = auto()
    T_CaseTestExpr = auto()
    T_ArrayExpr = auto()
    T_RowExpr = auto()
    T_RowCompareExpr = auto()
    T_CoalesceExpr = auto()
    T_MinMaxExpr = auto()
    T_SQLValueFunction = auto()
    T_XmlExpr = auto()
    T_NullTest = auto()
    T_BooleanTest = auto()
    T_CoerceToDomain = auto()
    T_CoerceToDomainValue = auto()
    T_SetToDefault = auto()
    T_CurrentOfExpr = auto()
    T_NextValueExpr = auto()
    T_InferenceElem = auto()
    T_TargetEntry = auto()
    T_RangeTblRef = auto()
    T_JoinExpr = auto()
    T_FromExpr = auto()
    T_OnConflictExpr = auto()
    T_IntoClause = auto()
    T_ExprState = auto()
    T_WindowFuncExprState = auto()
    T_SetExprState = auto()
    T_SubPlanState = auto()
    T_DomainConstraintState = auto()
    T_PlannerInfo = auto()
    T_PlannerGlobal = auto()
    T_RelOptInfo = auto()
    T_IndexOptInfo = auto()
    T_ForeignKeyOptInfo = auto()
    T_ParamPathInfo = auto()
    T_Path = auto()
    T_IndexPath = auto()
    T_BitmapHeapPath = auto()
    T_BitmapAndPath = auto()
    T_BitmapOrPath = auto()
    T_TidPath = auto()
    T_TidRangePath = auto()
    T_SubqueryScanPath = auto()
    T_ForeignPath = auto()
    T_CustomPath = auto()
    T_NestPath = auto()
    T_MergePath = auto()
    T_HashPath = auto()
    T_AppendPath = auto()
    T_MergeAppendPath = auto()
    T_GroupResultPath = auto()
    T_MaterialPath = auto()
    T_MemoizePath = auto()
    T_UniquePath = auto()
    T_GatherPath = auto()
    T_GatherMergePath = auto()
    T_ProjectionPath = auto()
    T_ProjectSetPath = auto()
    T_SortPath = auto()
    T_IncrementalSortPath = auto()
    T_GroupPath = auto()
    T_UpperUniquePath = auto()
    T_AggPath = auto()
    T_GroupingSetsPath = auto()
    T_MinMaxAggPath = auto()
    T_WindowAggPath = auto()
    T_SetOpPath = auto()
    T_RecursiveUnionPath = auto()
    T_LockRowsPath = auto()
    T_ModifyTablePath = auto()
    T_LimitPath = auto()
    T_EquivalenceClass = auto()
    T_EquivalenceMember = auto()
    T_PathKey = auto()
    T_PathKeyInfo = auto()
    T_PathTarget = auto()
    T_RestrictInfo = auto()
    T_IndexClause = auto()
    T_PlaceHolderVar = auto()
    T_SpecialJoinInfo = auto()
    T_AppendRelInfo = auto()
    T_RowIdentityVarInfo = auto()
    T_PlaceHolderInfo = auto()
    T_MinMaxAggInfo = auto()
    T_PlannerParamItem = auto()
    T_RollupData = auto()
    T_GroupingSetData = auto()
    T_StatisticExtInfo = auto()
    T_MergeAction = auto()
    T_AllocSetContext = auto()
    T_SlabContext = auto()
    T_GenerationContext = auto()
    T_Integer = auto()
    T_Float = auto()
    T_Boolean = auto()
    T_String = auto()
    T_BitString = auto()
    T_List = auto()
    T_IntList = auto()
    T_OidList = auto()
    T_ExtensibleNode = auto()
    T_RawStmt = auto()
    T_Query = auto()
    T_PlannedStmt = auto()
    T_InsertStmt = auto()
    T_DeleteStmt = auto()
    T_UpdateStmt = auto()
    T_MergeStmt = auto()
    T_SelectStmt = auto()
    T_ReturnStmt = auto()
    T_PLAssignStmt = auto()
    T_AlterTableStmt = auto()
    T_AlterTableCmd = auto()
    T_AlterDomainStmt = auto()
    T_SetOperationStmt = auto()
    T_GrantStmt = auto()
    T_GrantRoleStmt = auto()
    T_AlterDefaultPrivilegesStmt = auto()
    T_ClosePortalStmt = auto()
    T_ClusterStmt = auto()
    T_CopyStmt = auto()
    T_CreateStmt = auto()
    T_DefineStmt = auto()
    T_DropStmt = auto()
    T_TruncateStmt = auto()
    T_CommentStmt = auto()
    T_FetchStmt = auto()
    T_IndexStmt = auto()
    T_CreateFunctionStmt = auto()
    T_AlterFunctionStmt = auto()
    T_DoStmt = auto()
    T_RenameStmt = auto()
    T_RuleStmt = auto()
    T_NotifyStmt = auto()
    T_ListenStmt = auto()
    T_UnlistenStmt = auto()
    T_TransactionStmt = auto()
    T_ViewStmt = auto()
    T_LoadStmt = auto()
    T_CreateDomainStmt = auto()
    T_CreatedbStmt = auto()
    T_DropdbStmt = auto()
    T_VacuumStmt = auto()
    T_ExplainStmt = auto()
    T_CreateTableAsStmt = auto()
    T_CreateSeqStmt = auto()
    T_AlterSeqStmt = auto()
    T_VariableSetStmt = auto()
    T_VariableShowStmt = auto()
    T_DiscardStmt = auto()
    T_CreateTrigStmt = auto()
    T_CreatePLangStmt = auto()
    T_CreateRoleStmt = auto()
    T_AlterRoleStmt = auto()
    T_DropRoleStmt = auto()
    T_LockStmt = auto()
    T_ConstraintsSetStmt = auto()
    T_ReindexStmt = auto()
    T_CheckPointStmt = auto()
    T_CreateSchemaStmt = auto()
    T_AlterDatabaseStmt = auto()
    T_AlterDatabaseRefreshCollStmt = auto()
    T_AlterDatabaseSetStmt = auto()
    T_AlterRoleSetStmt = auto()
    T_CreateConversionStmt = auto()
    T_CreateCastStmt = auto()
    T_CreateOpClassStmt = auto()
    T_CreateOpFamilyStmt = auto()
    T_AlterOpFamilyStmt = auto()
    T_PrepareStmt = auto()
    T_ExecuteStmt = auto()
    T_DeallocateStmt = auto()
    T_DeclareCursorStmt = auto()
    T_CreateTableSpaceStmt = auto()
    T_DropTableSpaceStmt = auto()
    T_AlterObjectDependsStmt = auto()
    T_AlterObjectSchemaStmt = auto()
    T_AlterOwnerStmt = auto()
    T_AlterOperatorStmt = auto()
    T_AlterTypeStmt = auto()
    T_DropOwnedStmt = auto()
    T_ReassignOwnedStmt = auto()
    T_CompositeTypeStmt = auto()
    T_CreateEnumStmt = auto()
    T_CreateRangeStmt = auto()
    T_AlterEnumStmt = auto()
    T_AlterTSDictionaryStmt = auto()
    T_AlterTSConfigurationStmt = auto()
    T_CreateFdwStmt = auto()
    T_AlterFdwStmt = auto()
    T_CreateForeignServerStmt = auto()
    T_AlterForeignServerStmt = auto()
    T_CreateUserMappingStmt = auto()
    T_AlterUserMappingStmt = auto()
    T_DropUserMappingStmt = auto()
    T_AlterTableSpaceOptionsStmt = auto()
    T_AlterTableMoveAllStmt = auto()
    T_SecLabelStmt = auto()
    T_CreateForeignTableStmt = auto()
    T_ImportForeignSchemaStmt = auto()
    T_CreateExtensionStmt = auto()
    T_AlterExtensionStmt = auto()
    T_AlterExtensionContentsStmt = auto()
    T_CreateEventTrigStmt = auto()
    T_AlterEventTrigStmt = auto()
    T_RefreshMatViewStmt = auto()
    T_ReplicaIdentityStmt = auto()
    T_AlterSystemStmt = auto()
    T_CreatePolicyStmt = auto()
    T_AlterPolicyStmt = auto()
    T_CreateTransformStmt = auto()
    T_CreateAmStmt = auto()
    T_CreatePublicationStmt = auto()
    T_AlterPublicationStmt = auto()
    T_CreateSubscriptionStmt = auto()
    T_AlterSubscriptionStmt = auto()
    T_DropSubscriptionStmt = auto()
    T_CreateStatsStmt = auto()
    T_AlterCollationStmt = auto()
    T_CallStmt = auto()
    T_AlterStatsStmt = auto()
    T_A_Expr = auto()
    T_ColumnRef = auto()
    T_ParamRef = auto()
    T_A_Const = auto()
    T_FuncCall = auto()
    T_A_Star = auto()
    T_A_Indices = auto()
    T_A_Indirection = auto()
    T_A_ArrayExpr = auto()
    T_ResTarget = auto()
    T_MultiAssignRef = auto()
    T_TypeCast = auto()
    T_CollateClause = auto()
    T_SortBy = auto()
    T_WindowDef = auto()
    T_RangeSubselect = auto()
    T_RangeFunction = auto()
    T_RangeTableSample = auto()
    T_RangeTableFunc = auto()
    T_RangeTableFuncCol = auto()
    T_TypeName = auto()
    T_ColumnDef = auto()
    T_IndexElem = auto()
    T_StatsElem = auto()
    T_Constraint = auto()
    T_DefElem = auto()
    T_RangeTblEntry = auto()
    T_RangeTblFunction = auto()
    T_TableSampleClause = auto()
    T_WithCheckOption = auto()
    T_SortGroupClause = auto()
    T_GroupingSet = auto()
    T_WindowClause = auto()
    T_ObjectWithArgs = auto()
    T_AccessPriv = auto()
    T_CreateOpClassItem = auto()
    T_TableLikeClause = auto()
    T_FunctionParameter = auto()
    T_LockingClause = auto()
    T_RowMarkClause = auto()
    T_XmlSerialize = auto()
    T_WithClause = auto()
    T_InferClause = auto()
    T_OnConflictClause = auto()
    T_CTESearchClause = auto()
    T_CTECycleClause = auto()
    T_CommonTableExpr = auto()
    T_MergeWhenClause = auto()
    T_RoleSpec = auto()
    T_TriggerTransition = auto()
    T_PartitionElem = auto()
    T_PartitionSpec = auto()
    T_PartitionBoundSpec = auto()
    T_PartitionRangeDatum = auto()
    T_PartitionCmd = auto()
    T_VacuumRelation = auto()
    T_PublicationObjSpec = auto()
    T_PublicationTable = auto()
    T_IdentifySystemCmd = auto()
    T_BaseBackupCmd = auto()
    T_CreateReplicationSlotCmd = auto()
    T_DropReplicationSlotCmd = auto()
    T_ReadReplicationSlotCmd = auto()
    T_StartReplicationCmd = auto()
    T_TimeLineHistoryCmd = auto()
    T_TriggerData = auto()
    T_EventTriggerData = auto()
    T_ReturnSetInfo = auto()
    T_WindowObjectData = auto()
    T_TIDBitmap = auto()
    T_InlineCodeBlock = auto()
    T_FdwRoutine = auto()
    T_IndexAmRoutine = auto()
    T_TableAmRoutine = auto()
    T_TsmRoutine = auto()
    T_ForeignKeyCacheInfo = auto()
    T_CallContext = auto()
    T_SupportRequestSimplify = auto()
    T_SupportRequestSelectivity = auto()
    T_SupportRequestCost = auto()
    T_SupportRequestRows = auto()
    T_SupportRequestIndexCondition = auto()
    T_SupportRequestWFuncMonotonic = auto()

class OnConflictAction(IntEnum):
    ONCONFLICT_NONE = 0
    ONCONFLICT_NOTHING = auto()
    ONCONFLICT_UPDATE = auto()

class SetOpCmd(IntEnum):
    SETOPCMD_INTERSECT = 0
    SETOPCMD_INTERSECT_ALL = auto()
    SETOPCMD_EXCEPT = auto()
    SETOPCMD_EXCEPT_ALL = auto()

class SetOpStrategy(IntEnum):
    SETOP_SORTED = 0
    SETOP_HASHED = auto()


# #define-ed constants

AGGSPLITOP_COMBINE = 0x01

AGGSPLITOP_SKIPFINAL = 0x02

AGGSPLITOP_SERIALIZE = 0x04

AGGSPLITOP_DESERIALIZE = 0x08



# -*- coding: utf-8 -*-
# :Project:   pglast -- Enums used by the nodes
# :Created:   gio 03 ago 2017 17:08:26 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017, 2018, 2019, 2021 Lele Gaifax
#

# include/catalog
from .pg_am import *            # noqa
from .pg_attribute import *     # noqa
from .pg_class import *         # noqa
from .pg_trigger import *       # noqa

# include/nodes
from .lockoptions import *      # noqa
from .nodes import *            # noqa
from .parsenodes import *       # noqa
from .primnodes import *        # noqa

# include/storage
from .lockdefs import *         # noqa

# include/utils
from .xml import *              # noqa



# -*- coding: utf-8 -*-
# :Project:   pglast -- Cython interface with libpg_query
# :Created:   mer 02 ago 2017 15:12:49 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017, 2018, 2019, 2021, 2022 Lele Gaifax
#

#cython: language_level=3

from cpython.bytes cimport PyBytes_AsStringAndSize, PyBytes_FromStringAndSize
from cpython.list cimport PyList_New, PyList_SET_ITEM
from libc.stdint cimport int32_t, uint64_t, uint8_t
from libc cimport limits

from collections import namedtuple

from . import Error
from . cimport structs


include "ast.pyx"


class ParseError(Error):
    "Exception representing the error state returned by the PG parser."

    def __str__(self):
        message = self.args[0]
        if len(self.args) > 1:
            location = self.args[1]
            if location is not None:
                message += f', at index {location}'
        return message


class DeparseError(Error):
    "Exception representing the error state returned by the PG deparser."

    def __str__(self):
        message = self.args[0]
        if len(self.args) > 1:
            location = self.args[1]
            if location is not None:
                message += f', at position {location}'
        return message


cdef extern from "pg_query.h" nogil:

    int PG_VERSION_NUM

    ctypedef struct PgQueryError:
        char* message
        int lineno
        int cursorpos

    ctypedef struct PgQueryParseResult:
        char* parse_tree
        PgQueryError *error

    ctypedef struct PgQueryProtobuf:
        unsigned int len
        char* data

    ctypedef struct PgQueryProtobufParseResult:
        PgQueryProtobuf parse_tree
        PgQueryError* error;

    ctypedef struct PgQueryPlpgsqlParseResult:
        char* plpgsql_funcs
        PgQueryError* error

    ctypedef struct PgQueryFingerprintResult:
        char* fingerprint_str
        char* stderr_buffer
        PgQueryError* error

    ctypedef struct PgQuerySplitStmt:
        int stmt_location
        int stmt_len

    ctypedef struct PgQuerySplitResult:
        PgQuerySplitStmt** stmts
        int n_stmts
        char* stderr_buffer
        PgQueryError* error

    ctypedef struct PgQueryDeparseResult:
        char* query
        PgQueryError* error

    ctypedef struct PgQueryScanResult:
        PgQueryProtobuf pbuf
        PgQueryError *error

    PgQueryParseResult pg_query_parse(const char* input)
    void pg_query_free_parse_result(PgQueryParseResult result)

    PgQueryProtobufParseResult pg_query_parse_protobuf(const char* input)
    void pg_query_free_protobuf_parse_result(PgQueryProtobufParseResult result)

    PgQueryPlpgsqlParseResult pg_query_parse_plpgsql(const char* input)
    void pg_query_free_plpgsql_parse_result(PgQueryPlpgsqlParseResult result)

    PgQueryFingerprintResult pg_query_fingerprint(const char* input)
    void pg_query_free_fingerprint_result(PgQueryFingerprintResult result)

    PgQueryDeparseResult pg_query_deparse_protobuf(PgQueryProtobuf parse_tree)
    void pg_query_free_deparse_result(PgQueryDeparseResult result)

    PgQuerySplitResult pg_query_split_with_scanner(const char* input)
    PgQuerySplitResult pg_query_split_with_parser(const char* input)
    void pg_query_free_split_result(PgQuerySplitResult result)

    PgQueryScanResult pg_query_scan(const char* input)
    void pg_query_free_scan_result(PgQueryScanResult result)


cdef extern from "src/pg_query_internal.h" nogil:
    ctypedef struct PgQueryInternalParsetreeAndError:
        structs.List* tree
        char* stderr_buffer
        PgQueryError* error

    ctypedef void* MemoryContext

    MemoryContext pg_query_enter_memory_context()
    void pg_query_exit_memory_context(MemoryContext ctx)

    PgQueryInternalParsetreeAndError pg_query_raw_parse(const char* input)


cdef extern from "protobuf-c/protobuf-c.h":
    ctypedef struct ProtobufCEnumDescriptor:
        pass

    ctypedef struct ProtobufCEnumValue:
        const char* name
        const char* c_name
        int value

    ProtobufCEnumValue* protobuf_c_enum_descriptor_get_value(const ProtobufCEnumDescriptor* d,
                                                             int value)


cdef extern from "protobuf/pg_query.pb-c.h" nogil:
    ctypedef enum PgQuery__Token:
        pass

    ctypedef enum PgQuery__KeywordKind:
        pass

    ctypedef struct PgQuery__ScanToken:
        int32_t start
        int32_t end
        PgQuery__Token token
        PgQuery__KeywordKind keyword_kind

    ctypedef struct PgQuery__ScanResult:
        size_t n_tokens
        PgQuery__ScanToken **tokens

    ProtobufCEnumDescriptor pg_query__token__descriptor
    ProtobufCEnumDescriptor pg_query__keyword_kind__descriptor

    PgQuery__ScanResult* pg_query__scan_result__unpack(void* allocator, size_t len,
                                                       const uint8_t* data)
    void pg_query__scan_result__free_unpacked(PgQuery__ScanResult* message, void* allocator)


LONG_MAX = limits.LONG_MAX


cdef class Displacements:
    """
    Helper class used to find the index of Unicode character from its offset in the
    corresponding UTF-8 encoded array.

    Example:

      >>> unicode = '€ 0.01'
      >>> utf8 = unicode.encode('utf-8')
      >>> d = pglast.parser.Displacements(unicode)
      >>> for offset in range(len(utf8)):
      ...   print(f'{offset} [{utf8[offset]:2x}] -> {d(offset)} [{unicode[d(offset)]}]')
      ...
      0 [e2] -> 0 [€]
      1 [82] -> 0 [€]
      2 [ac] -> 0 [€]
      3 [20] -> 1 [ ]
      4 [30] -> 2 [0]
      5 [2e] -> 3 [.]
      6 [30] -> 4 [0]
      7 [31] -> 5 [1]
    """

    # Build a sequence of (offset, index) tuples, for each character in the original string
    # that requires more than one byte once encoded in UTF-8. The sequence is kept in reverse
    # order, because to compute the index for an arbitrary offset we simply find the first
    # tuple with an offset less than or equal the given offset.

    cdef tuple displacements
    cdef unsigned long max_offset

    def __init__(self, str s):
        cdef unsigned long idx = 0
        cdef unsigned long ofs = 0
        cdef size_t c_len_in_utf8

        disps = []
        for c in s:
            c_len_in_utf8 = len(c.encode('utf-8'))
            if c_len_in_utf8 > 1:
                while c_len_in_utf8 > 0:
                    disps.append((ofs, idx))
                    c_len_in_utf8 -= 1
                    ofs += 1
            else:
                ofs += 1
            idx += 1
        self.max_offset = ofs - 1
        self.displacements = tuple(reversed(disps))

    def __call__(self, offset):
        if not 0 <= offset <= self.max_offset:
            # It may happen that the error is actually at the very end of the statement, giving
            # the message "syntax error at end of input", with a position that is beyond the
            # end of the string. Return None in such case
            return None

        for o, i in self.displacements:
            if o <= offset:
                return i + (offset - o)

        return offset


def get_postgresql_version():
    "Return the ``PostgreSQL`` version as a tuple (`major`, `minor`)."

    version = PG_VERSION_NUM
    major, minor = divmod(version, 10_000)
    return (major, minor)


def parse_sql(str query):
    "Parse the given ``SQL`` `query` and return its abstract syntax tree."

    cdef PgQueryInternalParsetreeAndError parsed
    cdef MemoryContext mctx
    cdef int i
    cdef const char *cstring

    utf8 = query.encode('utf-8')
    offset_to_index = Displacements(query)
    cstring = utf8

    mctx = pg_query_enter_memory_context()

    with nogil:
        parsed = pg_query_raw_parse(cstring)

    try:
        if parsed.tree is not NULL:
            stmts = PyTuple_New(parsed.tree.length)
            for i in range(parsed.tree.length):
                item = create(structs.list_nth(parsed.tree, i), offset_to_index)
                Py_INCREF(item)
                PyTuple_SET_ITEM(stmts, i, item)
            return stmts
        elif parsed.error is NULL:
            return ()
        else:
            message = parsed.error.message.decode('utf8')
            raise ParseError(message, offset_to_index(parsed.error.cursorpos-1))
    finally:
        pg_query_exit_memory_context(mctx);


def parse_sql_json(str query):
    "Parse the given ``SQL`` `query` and return its ``JSON`` encoded syntax tree."

    cdef PgQueryParseResult parsed
    cdef const char *cstring

    utf8 = query.encode('utf-8')
    cstring = utf8

    with nogil:
        parsed = pg_query_parse(cstring)

    try:
        if parsed.error:
            message = parsed.error.message.decode('utf8')
            offset_to_index = Displacements(query)
            raise ParseError(message, offset_to_index(parsed.error.cursorpos-1))

        return parsed.parse_tree.decode('utf8')
    finally:
        with nogil:
            pg_query_free_parse_result(parsed)


def parse_sql_protobuf(str query):
    "Parse the given ``SQL`` `query` and return its ``protobuf`` encoded syntax tree."

    cdef PgQueryProtobufParseResult parsed
    cdef const char *cstring

    utf8 = query.encode('utf-8')
    cstring = utf8

    with nogil:
        parsed = pg_query_parse_protobuf(cstring)

    try:
        if parsed.error:
            message = parsed.error.message.decode('utf8')
            cursorpos = parsed.error.cursorpos
            offset_to_index = Displacements(query)
            raise ParseError(message, offset_to_index(parsed.error.cursorpos-1))

        return PyBytes_FromStringAndSize(parsed.parse_tree.data, parsed.parse_tree.len)
    finally:
        with nogil:
            pg_query_free_protobuf_parse_result(parsed)


def parse_plpgsql_json(str query):
    "Parse the given ``pgpgsql`` `query` and return its ``JSON`` encoded syntax tree."

    cdef PgQueryPlpgsqlParseResult parsed
    cdef const char *cstring

    utf8 = query.encode('utf-8')
    cstring = utf8

    with nogil:
        parsed = pg_query_parse_plpgsql(cstring)

    try:
        if parsed.error:
            message = parsed.error.message.decode('utf8')
            offset_to_index = Displacements(query)
            raise ParseError(message, offset_to_index(parsed.error.cursorpos-1))

        return parsed.plpgsql_funcs.decode('utf8')
    finally:
        with nogil:
            pg_query_free_plpgsql_parse_result(parsed)


def fingerprint(str query):
    "Compute and return a *signature* of the given ``SQL`` `query`."

    cdef PgQueryFingerprintResult result
    cdef const char *cstring

    utf8 = query.encode('utf-8')
    cstring = utf8

    with nogil:
        result = pg_query_fingerprint(cstring)

    try:
        if result.error:
            message = result.error.message.decode('utf8')
            offset_to_index = Displacements(query)
            raise ParseError(message, offset_to_index(result.error.cursorpos-1))

        return result.fingerprint_str.decode('ascii')
    finally:
        with nogil:
            pg_query_free_fingerprint_result(result)


def split(str stmts, bint with_parser=True, bint only_slices=False):
    """Split the given `stmts` string into a sequence of the single ``SQL`` statements.

    By default this uses the *parser* to perform the job; when `with_parser` is ``False``
    the *scanner* variant is used, indicated when the statements may contain parse errors.

    When `only_slices` is ``True``, return a sequence of ``slice`` instances, one for each
    statement, instead of statements text.

    NB: leading and trailing whitespace are removed from the statements.
    """

    cdef PgQuerySplitResult splitted
    cdef const char *cstring
    cdef int i = 0
    cdef int prev_offset = 0
    cdef int start
    cdef int end

    utf8 = stmts.encode('utf-8')
    cstring = utf8

    with nogil:
        if with_parser:
            splitted = pg_query_split_with_parser(cstring)
        else:
            splitted = pg_query_split_with_scanner(cstring)

    try:
        if splitted.error:
            message = splitted.error.message.decode('utf8')
            offset_to_index = Displacements(stmts)
            raise ParseError(message, offset_to_index(splitted.error.cursorpos-1))

        result = []
        while i < splitted.n_stmts:
            start = splitted.stmts[i].stmt_location
            end = splitted.stmts[i].stmt_location + splitted.stmts[i].stmt_len
            stmt = utf8[start:end].decode('utf-8').strip()
            if only_slices:
                # Adjust offsets, we remove leading/trailing whitespace above
                cur_offset = stmts.index(stmt, prev_offset)
                result.append(slice(cur_offset, cur_offset + len(stmt)))
                prev_offset = cur_offset + 1
            else:
                result.append(stmt)
            i += 1
        return tuple(result)
    finally:
        with nogil:
            pg_query_free_split_result(splitted)


def deparse_protobuf(bytes protobuf):
    "Convert the ``protobuf`` serialized parse tree into an equivalent ``SQL`` statement."

    cdef PgQueryProtobuf tree
    cdef PgQueryDeparseResult deparsed

    PyBytes_AsStringAndSize(protobuf, &tree.data, <Py_ssize_t *>&tree.len)
    with nogil:
         deparsed = pg_query_deparse_protobuf(tree)

    try:
        if deparsed.error:
            message = deparsed.error.message.decode('utf8')
            raise DeparseError(message, deparsed.error.cursorpos)

        return deparsed.query.decode('utf-8')
    finally:
        with nogil:
            pg_query_free_deparse_result(deparsed)


Token = namedtuple('Token', ('start', 'end', 'name', 'kind'))


def scan(str query):
    "Lexify the given ``SQL`` `query` and return a list of its lexical tokens."

    cdef PgQueryScanResult scanned
    cdef PgQuery__ScanResult* scan_result
    cdef PgQuery__ScanToken* scan_token
    cdef const ProtobufCEnumValue* tkind
    cdef const ProtobufCEnumValue* kwkind
    cdef const char* cstring
    cdef size_t i

    utf8 = query.encode('utf-8')
    offset_to_index = Displacements(query)
    cstring = utf8

    with nogil:
        scanned = pg_query_scan(cstring)

    try:
        if scanned.error:
            message = scanned.error.message.decode('utf8')
            raise ParseError(message, offset_to_index(scanned.error.cursorpos-1))

        with nogil:
            scan_result = pg_query__scan_result__unpack(NULL, scanned.pbuf.len,
                                                        <uint8_t*> scanned.pbuf.data)

        result = PyList_New(scan_result.n_tokens)

        for i in range(scan_result.n_tokens):
            scan_token = scan_result.tokens[i]
            tkind = protobuf_c_enum_descriptor_get_value(&pg_query__token__descriptor,
                                                         scan_token.token)
            kwkind = protobuf_c_enum_descriptor_get_value(&pg_query__keyword_kind__descriptor,
                                                          scan_token.keyword_kind)

            token = Token(offset_to_index(scan_token.start), offset_to_index(scan_token.end-1),
                          tkind.name.decode('ascii') if tkind != NULL else "UNKNOWN",
                          kwkind.name.decode('ascii'))
            Py_INCREF(token)
            PyList_SET_ITEM(result, i, token)

        with nogil:
            pg_query__scan_result__free_unpacked(scan_result, NULL)
    finally:
        with nogil:
            pg_query_free_scan_result(scanned)

    return result



# -*- coding: utf-8 -*-
# :Project:   pglast -- DO NOT EDIT: automatically extracted from parsenodes.h @ 15-4.0.0-0-g1ada550
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017-2022 Lele Gaifax
#

from enum import Enum, IntEnum, IntFlag, auto

try:
    from enum import StrEnum
except ImportError:  # pragma: no cover
    # Python < 3.10
    class StrEnum(str, Enum):
        pass


class A_Expr_Kind(IntEnum):
    AEXPR_OP = 0
    AEXPR_OP_ANY = auto()
    AEXPR_OP_ALL = auto()
    AEXPR_DISTINCT = auto()
    AEXPR_NOT_DISTINCT = auto()
    AEXPR_NULLIF = auto()
    AEXPR_IN = auto()
    AEXPR_LIKE = auto()
    AEXPR_ILIKE = auto()
    AEXPR_SIMILAR = auto()
    AEXPR_BETWEEN = auto()
    AEXPR_NOT_BETWEEN = auto()
    AEXPR_BETWEEN_SYM = auto()
    AEXPR_NOT_BETWEEN_SYM = auto()

class AlterPublicationAction(IntEnum):
    AP_AddObjects = 0
    AP_DropObjects = auto()
    AP_SetObjects = auto()

class AlterSubscriptionType(IntEnum):
    ALTER_SUBSCRIPTION_OPTIONS = 0
    ALTER_SUBSCRIPTION_CONNECTION = auto()
    ALTER_SUBSCRIPTION_SET_PUBLICATION = auto()
    ALTER_SUBSCRIPTION_ADD_PUBLICATION = auto()
    ALTER_SUBSCRIPTION_DROP_PUBLICATION = auto()
    ALTER_SUBSCRIPTION_REFRESH = auto()
    ALTER_SUBSCRIPTION_ENABLED = auto()
    ALTER_SUBSCRIPTION_SKIP = auto()

class AlterTSConfigType(IntEnum):
    ALTER_TSCONFIG_ADD_MAPPING = 0
    ALTER_TSCONFIG_ALTER_MAPPING_FOR_TOKEN = auto()
    ALTER_TSCONFIG_REPLACE_DICT = auto()
    ALTER_TSCONFIG_REPLACE_DICT_FOR_TOKEN = auto()
    ALTER_TSCONFIG_DROP_MAPPING = auto()

class AlterTableType(IntEnum):
    AT_AddColumn = 0
    AT_AddColumnRecurse = auto()
    AT_AddColumnToView = auto()
    AT_ColumnDefault = auto()
    AT_CookedColumnDefault = auto()
    AT_DropNotNull = auto()
    AT_SetNotNull = auto()
    AT_DropExpression = auto()
    AT_CheckNotNull = auto()
    AT_SetStatistics = auto()
    AT_SetOptions = auto()
    AT_ResetOptions = auto()
    AT_SetStorage = auto()
    AT_SetCompression = auto()
    AT_DropColumn = auto()
    AT_DropColumnRecurse = auto()
    AT_AddIndex = auto()
    AT_ReAddIndex = auto()
    AT_AddConstraint = auto()
    AT_AddConstraintRecurse = auto()
    AT_ReAddConstraint = auto()
    AT_ReAddDomainConstraint = auto()
    AT_AlterConstraint = auto()
    AT_ValidateConstraint = auto()
    AT_ValidateConstraintRecurse = auto()
    AT_AddIndexConstraint = auto()
    AT_DropConstraint = auto()
    AT_DropConstraintRecurse = auto()
    AT_ReAddComment = auto()
    AT_AlterColumnType = auto()
    AT_AlterColumnGenericOptions = auto()
    AT_ChangeOwner = auto()
    AT_ClusterOn = auto()
    AT_DropCluster = auto()
    AT_SetLogged = auto()
    AT_SetUnLogged = auto()
    AT_DropOids = auto()
    AT_SetAccessMethod = auto()
    AT_SetTableSpace = auto()
    AT_SetRelOptions = auto()
    AT_ResetRelOptions = auto()
    AT_ReplaceRelOptions = auto()
    AT_EnableTrig = auto()
    AT_EnableAlwaysTrig = auto()
    AT_EnableReplicaTrig = auto()
    AT_DisableTrig = auto()
    AT_EnableTrigAll = auto()
    AT_DisableTrigAll = auto()
    AT_EnableTrigUser = auto()
    AT_DisableTrigUser = auto()
    AT_EnableRule = auto()
    AT_EnableAlwaysRule = auto()
    AT_EnableReplicaRule = auto()
    AT_DisableRule = auto()
    AT_AddInherit = auto()
    AT_DropInherit = auto()
    AT_AddOf = auto()
    AT_DropOf = auto()
    AT_ReplicaIdentity = auto()
    AT_EnableRowSecurity = auto()
    AT_DisableRowSecurity = auto()
    AT_ForceRowSecurity = auto()
    AT_NoForceRowSecurity = auto()
    AT_GenericOptions = auto()
    AT_AttachPartition = auto()
    AT_DetachPartition = auto()
    AT_DetachPartitionFinalize = auto()
    AT_AddIdentity = auto()
    AT_SetIdentity = auto()
    AT_DropIdentity = auto()
    AT_ReAddStatistics = auto()

class CTEMaterialize(IntEnum):
    CTEMaterializeDefault = 0
    CTEMaterializeAlways = auto()
    CTEMaterializeNever = auto()

class ConstrType(IntEnum):
    CONSTR_NULL = 0
    CONSTR_NOTNULL = auto()
    CONSTR_DEFAULT = auto()
    CONSTR_IDENTITY = auto()
    CONSTR_GENERATED = auto()
    CONSTR_CHECK = auto()
    CONSTR_PRIMARY = auto()
    CONSTR_UNIQUE = auto()
    CONSTR_EXCLUSION = auto()
    CONSTR_FOREIGN = auto()
    CONSTR_ATTR_DEFERRABLE = auto()
    CONSTR_ATTR_NOT_DEFERRABLE = auto()
    CONSTR_ATTR_DEFERRED = auto()
    CONSTR_ATTR_IMMEDIATE = auto()

class DefElemAction(IntEnum):
    DEFELEM_UNSPEC = 0
    DEFELEM_SET = auto()
    DEFELEM_ADD = auto()
    DEFELEM_DROP = auto()

class DiscardMode(IntEnum):
    DISCARD_ALL = 0
    DISCARD_PLANS = auto()
    DISCARD_SEQUENCES = auto()
    DISCARD_TEMP = auto()

class DropBehavior(IntEnum):
    DROP_RESTRICT = 0
    DROP_CASCADE = auto()

class FetchDirection(IntEnum):
    FETCH_FORWARD = 0
    FETCH_BACKWARD = auto()
    FETCH_ABSOLUTE = auto()
    FETCH_RELATIVE = auto()

class FunctionParameterMode(StrEnum):
    FUNC_PARAM_IN = 'i'
    FUNC_PARAM_OUT = 'o'
    FUNC_PARAM_INOUT = 'b'
    FUNC_PARAM_VARIADIC = 'v'
    FUNC_PARAM_TABLE = 't'
    FUNC_PARAM_DEFAULT = 'd'

class GrantTargetType(IntEnum):
    ACL_TARGET_OBJECT = 0
    ACL_TARGET_ALL_IN_SCHEMA = auto()
    ACL_TARGET_DEFAULTS = auto()

class GroupingSetKind(IntEnum):
    GROUPING_SET_EMPTY = 0
    GROUPING_SET_SIMPLE = auto()
    GROUPING_SET_ROLLUP = auto()
    GROUPING_SET_CUBE = auto()
    GROUPING_SET_SETS = auto()

class ImportForeignSchemaType(IntEnum):
    FDW_IMPORT_SCHEMA_ALL = 0
    FDW_IMPORT_SCHEMA_LIMIT_TO = auto()
    FDW_IMPORT_SCHEMA_EXCEPT = auto()

class ObjectType(IntEnum):
    OBJECT_ACCESS_METHOD = 0
    OBJECT_AGGREGATE = auto()
    OBJECT_AMOP = auto()
    OBJECT_AMPROC = auto()
    OBJECT_ATTRIBUTE = auto()
    OBJECT_CAST = auto()
    OBJECT_COLUMN = auto()
    OBJECT_COLLATION = auto()
    OBJECT_CONVERSION = auto()
    OBJECT_DATABASE = auto()
    OBJECT_DEFAULT = auto()
    OBJECT_DEFACL = auto()
    OBJECT_DOMAIN = auto()
    OBJECT_DOMCONSTRAINT = auto()
    OBJECT_EVENT_TRIGGER = auto()
    OBJECT_EXTENSION = auto()
    OBJECT_FDW = auto()
    OBJECT_FOREIGN_SERVER = auto()
    OBJECT_FOREIGN_TABLE = auto()
    OBJECT_FUNCTION = auto()
    OBJECT_INDEX = auto()
    OBJECT_LANGUAGE = auto()
    OBJECT_LARGEOBJECT = auto()
    OBJECT_MATVIEW = auto()
    OBJECT_OPCLASS = auto()
    OBJECT_OPERATOR = auto()
    OBJECT_OPFAMILY = auto()
    OBJECT_PARAMETER_ACL = auto()
    OBJECT_POLICY = auto()
    OBJECT_PROCEDURE = auto()
    OBJECT_PUBLICATION = auto()
    OBJECT_PUBLICATION_NAMESPACE = auto()
    OBJECT_PUBLICATION_REL = auto()
    OBJECT_ROLE = auto()
    OBJECT_ROUTINE = auto()
    OBJECT_RULE = auto()
    OBJECT_SCHEMA = auto()
    OBJECT_SEQUENCE = auto()
    OBJECT_SUBSCRIPTION = auto()
    OBJECT_STATISTIC_EXT = auto()
    OBJECT_TABCONSTRAINT = auto()
    OBJECT_TABLE = auto()
    OBJECT_TABLESPACE = auto()
    OBJECT_TRANSFORM = auto()
    OBJECT_TRIGGER = auto()
    OBJECT_TSCONFIGURATION = auto()
    OBJECT_TSDICTIONARY = auto()
    OBJECT_TSPARSER = auto()
    OBJECT_TSTEMPLATE = auto()
    OBJECT_TYPE = auto()
    OBJECT_USER_MAPPING = auto()
    OBJECT_VIEW = auto()

class OverridingKind(IntEnum):
    OVERRIDING_NOT_SET = 0
    OVERRIDING_USER_VALUE = auto()
    OVERRIDING_SYSTEM_VALUE = auto()

class PartitionRangeDatumKind(IntEnum):
    PARTITION_RANGE_DATUM_MINVALUE = -1
    PARTITION_RANGE_DATUM_VALUE = 0
    PARTITION_RANGE_DATUM_MAXVALUE = 1

class PublicationObjSpecType(IntEnum):
    PUBLICATIONOBJ_TABLE = 0
    PUBLICATIONOBJ_TABLES_IN_SCHEMA = auto()
    PUBLICATIONOBJ_TABLES_IN_CUR_SCHEMA = auto()
    PUBLICATIONOBJ_CONTINUATION = auto()

class QuerySource(IntEnum):
    QSRC_ORIGINAL = 0
    QSRC_PARSER = auto()
    QSRC_INSTEAD_RULE = auto()
    QSRC_QUAL_INSTEAD_RULE = auto()
    QSRC_NON_INSTEAD_RULE = auto()

class RTEKind(IntEnum):
    RTE_RELATION = 0
    RTE_SUBQUERY = auto()
    RTE_JOIN = auto()
    RTE_FUNCTION = auto()
    RTE_TABLEFUNC = auto()
    RTE_VALUES = auto()
    RTE_CTE = auto()
    RTE_NAMEDTUPLESTORE = auto()
    RTE_RESULT = auto()

class ReindexObjectType(IntEnum):
    REINDEX_OBJECT_INDEX = 0
    REINDEX_OBJECT_TABLE = auto()
    REINDEX_OBJECT_SCHEMA = auto()
    REINDEX_OBJECT_SYSTEM = auto()
    REINDEX_OBJECT_DATABASE = auto()

class RoleSpecType(IntEnum):
    ROLESPEC_CSTRING = 0
    ROLESPEC_CURRENT_ROLE = auto()
    ROLESPEC_CURRENT_USER = auto()
    ROLESPEC_SESSION_USER = auto()
    ROLESPEC_PUBLIC = auto()

class RoleStmtType(IntEnum):
    ROLESTMT_ROLE = 0
    ROLESTMT_USER = auto()
    ROLESTMT_GROUP = auto()

class SetOperation(IntEnum):
    SETOP_NONE = 0
    SETOP_UNION = auto()
    SETOP_INTERSECT = auto()
    SETOP_EXCEPT = auto()

class SetQuantifier(IntEnum):
    SET_QUANTIFIER_DEFAULT = 0
    SET_QUANTIFIER_ALL = auto()
    SET_QUANTIFIER_DISTINCT = auto()

class SortByDir(IntEnum):
    SORTBY_DEFAULT = 0
    SORTBY_ASC = auto()
    SORTBY_DESC = auto()
    SORTBY_USING = auto()

class SortByNulls(IntEnum):
    SORTBY_NULLS_DEFAULT = 0
    SORTBY_NULLS_FIRST = auto()
    SORTBY_NULLS_LAST = auto()

class TableLikeOption(IntFlag):
    CREATE_TABLE_LIKE_COMMENTS = 1 << 0
    CREATE_TABLE_LIKE_COMPRESSION = 1 << 1
    CREATE_TABLE_LIKE_CONSTRAINTS = 1 << 2
    CREATE_TABLE_LIKE_DEFAULTS = 1 << 3
    CREATE_TABLE_LIKE_GENERATED = 1 << 4
    CREATE_TABLE_LIKE_IDENTITY = 1 << 5
    CREATE_TABLE_LIKE_INDEXES = 1 << 6
    CREATE_TABLE_LIKE_STATISTICS = 1 << 7
    CREATE_TABLE_LIKE_STORAGE = 1 << 8
    CREATE_TABLE_LIKE_ALL = 0x7FFFFFFF

class TransactionStmtKind(IntEnum):
    TRANS_STMT_BEGIN = 0
    TRANS_STMT_START = auto()
    TRANS_STMT_COMMIT = auto()
    TRANS_STMT_ROLLBACK = auto()
    TRANS_STMT_SAVEPOINT = auto()
    TRANS_STMT_RELEASE = auto()
    TRANS_STMT_ROLLBACK_TO = auto()
    TRANS_STMT_PREPARE = auto()
    TRANS_STMT_COMMIT_PREPARED = auto()
    TRANS_STMT_ROLLBACK_PREPARED = auto()

class VariableSetKind(IntEnum):
    VAR_SET_VALUE = 0
    VAR_SET_DEFAULT = auto()
    VAR_SET_CURRENT = auto()
    VAR_SET_MULTI = auto()
    VAR_RESET = auto()
    VAR_RESET_ALL = auto()

class ViewCheckOption(IntEnum):
    NO_CHECK_OPTION = 0
    LOCAL_CHECK_OPTION = auto()
    CASCADED_CHECK_OPTION = auto()

class WCOKind(IntEnum):
    WCO_VIEW_CHECK = 0
    WCO_RLS_INSERT_CHECK = auto()
    WCO_RLS_UPDATE_CHECK = auto()
    WCO_RLS_CONFLICT_CHECK = auto()
    WCO_RLS_MERGE_UPDATE_CHECK = auto()
    WCO_RLS_MERGE_DELETE_CHECK = auto()


# #define-ed constants

ACL_INSERT = 1<<0

ACL_SELECT = 1<<1

ACL_UPDATE = 1<<2

ACL_DELETE = 1<<3

ACL_TRUNCATE = 1<<4

ACL_REFERENCES = 1<<5

ACL_TRIGGER = 1<<6

ACL_EXECUTE = 1<<7

ACL_USAGE = 1<<8

ACL_CREATE = 1<<9

ACL_CREATE_TEMP = 1<<10

ACL_CONNECT = 1<<11

ACL_SET = 1<<12

ACL_ALTER_SYSTEM = 1<<13

N_ACL_RIGHTS = 14

ACL_NO_RIGHTS = 0

FRAMEOPTION_NONDEFAULT = 0x00001

FRAMEOPTION_RANGE = 0x00002

FRAMEOPTION_ROWS = 0x00004

FRAMEOPTION_GROUPS = 0x00008

FRAMEOPTION_BETWEEN = 0x00010

FRAMEOPTION_START_UNBOUNDED_PRECEDING = 0x00020

FRAMEOPTION_END_UNBOUNDED_PRECEDING = 0x00040

FRAMEOPTION_START_UNBOUNDED_FOLLOWING = 0x00080

FRAMEOPTION_END_UNBOUNDED_FOLLOWING = 0x00100

FRAMEOPTION_START_CURRENT_ROW = 0x00200

FRAMEOPTION_END_CURRENT_ROW = 0x00400

FRAMEOPTION_START_OFFSET_PRECEDING = 0x00800

FRAMEOPTION_END_OFFSET_PRECEDING = 0x01000

FRAMEOPTION_START_OFFSET_FOLLOWING = 0x02000

FRAMEOPTION_END_OFFSET_FOLLOWING = 0x04000

FRAMEOPTION_EXCLUDE_CURRENT_ROW = 0x08000

FRAMEOPTION_EXCLUDE_GROUP = 0x10000

FRAMEOPTION_EXCLUDE_TIES = 0x20000

PARTITION_STRATEGY_HASH = 'h'

PARTITION_STRATEGY_LIST = 'l'

PARTITION_STRATEGY_RANGE = 'r'

FKCONSTR_ACTION_NOACTION = 'a'

FKCONSTR_ACTION_RESTRICT = 'r'

FKCONSTR_ACTION_CASCADE = 'c'

FKCONSTR_ACTION_SETNULL = 'n'

FKCONSTR_ACTION_SETDEFAULT = 'd'

FKCONSTR_MATCH_FULL = 'f'

FKCONSTR_MATCH_PARTIAL = 'p'

FKCONSTR_MATCH_SIMPLE = 's'

OPCLASS_ITEM_OPERATOR = 1

OPCLASS_ITEM_FUNCTION = 2

OPCLASS_ITEM_STORAGETYPE = 3

CURSOR_OPT_BINARY = 0x0001

CURSOR_OPT_SCROLL = 0x0002

CURSOR_OPT_NO_SCROLL = 0x0004

CURSOR_OPT_INSENSITIVE = 0x0008

CURSOR_OPT_ASENSITIVE = 0x0010

CURSOR_OPT_HOLD = 0x0020

CURSOR_OPT_FAST_PLAN = 0x0100

CURSOR_OPT_GENERIC_PLAN = 0x0200

CURSOR_OPT_CUSTOM_PLAN = 0x0400

CURSOR_OPT_PARALLEL_OK = 0x0800

FETCH_ALL = 9223372036854775807



# -*- coding: utf-8 -*-
# :Project:   pglast -- DO NOT EDIT: automatically extracted from primnodes.h @ 15-4.0.0-0-g1ada550
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017-2022 Lele Gaifax
#

from enum import Enum, IntEnum, IntFlag, auto

try:
    from enum import StrEnum
except ImportError:  # pragma: no cover
    # Python < 3.10
    class StrEnum(str, Enum):
        pass


class BoolExprType(IntEnum):
    AND_EXPR = 0
    OR_EXPR = auto()
    NOT_EXPR = auto()

class BoolTestType(IntEnum):
    IS_TRUE = 0
    IS_NOT_TRUE = auto()
    IS_FALSE = auto()
    IS_NOT_FALSE = auto()
    IS_UNKNOWN = auto()
    IS_NOT_UNKNOWN = auto()

class CoercionContext(IntEnum):
    COERCION_IMPLICIT = 0
    COERCION_ASSIGNMENT = auto()
    COERCION_PLPGSQL = auto()
    COERCION_EXPLICIT = auto()

class CoercionForm(IntEnum):
    COERCE_EXPLICIT_CALL = 0
    COERCE_EXPLICIT_CAST = auto()
    COERCE_IMPLICIT_CAST = auto()
    COERCE_SQL_SYNTAX = auto()

class MinMaxOp(IntEnum):
    IS_GREATEST = 0
    IS_LEAST = auto()

class NullTestType(IntEnum):
    IS_NULL = 0
    IS_NOT_NULL = auto()

class OnCommitAction(IntEnum):
    ONCOMMIT_NOOP = 0
    ONCOMMIT_PRESERVE_ROWS = auto()
    ONCOMMIT_DELETE_ROWS = auto()
    ONCOMMIT_DROP = auto()

class ParamKind(IntEnum):
    PARAM_EXTERN = 0
    PARAM_EXEC = auto()
    PARAM_SUBLINK = auto()
    PARAM_MULTIEXPR = auto()

class RowCompareType(IntEnum):
    ROWCOMPARE_LT = 1
    ROWCOMPARE_LE = 2
    ROWCOMPARE_EQ = 3
    ROWCOMPARE_GE = 4
    ROWCOMPARE_GT = 5
    ROWCOMPARE_NE = 6

class SQLValueFunctionOp(IntEnum):
    SVFOP_CURRENT_DATE = 0
    SVFOP_CURRENT_TIME = auto()
    SVFOP_CURRENT_TIME_N = auto()
    SVFOP_CURRENT_TIMESTAMP = auto()
    SVFOP_CURRENT_TIMESTAMP_N = auto()
    SVFOP_LOCALTIME = auto()
    SVFOP_LOCALTIME_N = auto()
    SVFOP_LOCALTIMESTAMP = auto()
    SVFOP_LOCALTIMESTAMP_N = auto()
    SVFOP_CURRENT_ROLE = auto()
    SVFOP_CURRENT_USER = auto()
    SVFOP_USER = auto()
    SVFOP_SESSION_USER = auto()
    SVFOP_CURRENT_CATALOG = auto()
    SVFOP_CURRENT_SCHEMA = auto()

class SubLinkType(IntEnum):
    EXISTS_SUBLINK = 0
    ALL_SUBLINK = auto()
    ANY_SUBLINK = auto()
    ROWCOMPARE_SUBLINK = auto()
    EXPR_SUBLINK = auto()
    MULTIEXPR_SUBLINK = auto()
    ARRAY_SUBLINK = auto()
    CTE_SUBLINK = auto()

class XmlExprOp(IntEnum):
    IS_XMLCONCAT = 0
    IS_XMLELEMENT = auto()
    IS_XMLFOREST = auto()
    IS_XMLPARSE = auto()
    IS_XMLPI = auto()
    IS_XMLROOT = auto()
    IS_XMLSERIALIZE = auto()
    IS_DOCUMENT = auto()

class XmlOptionType(IntEnum):
    XMLOPTION_DOCUMENT = 0
    XMLOPTION_CONTENT = auto()



# -*- coding: utf-8 -*-
# :Project:   pglast -- DO NOT EDIT: automatically extracted from lockoptions.h @ 15-4.0.0-0-g1ada550
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017-2022 Lele Gaifax
#

from enum import Enum, IntEnum, IntFlag, auto

try:
    from enum import StrEnum
except ImportError:  # pragma: no cover
    # Python < 3.10
    class StrEnum(str, Enum):
        pass


class LockClauseStrength(IntEnum):
    LCS_NONE = 0
    LCS_FORKEYSHARE = auto()
    LCS_FORSHARE = auto()
    LCS_FORNOKEYUPDATE = auto()
    LCS_FORUPDATE = auto()

class LockTupleMode(IntEnum):
    LockTupleKeyShare = 0
    LockTupleShare = auto()
    LockTupleNoKeyExclusive = auto()
    LockTupleExclusive = auto()

class LockWaitPolicy(IntEnum):
    LockWaitBlock = 0
    LockWaitSkip = auto()
    LockWaitError = auto()



# -*- coding: utf-8 -*-
# :Project:   pglast -- Tests on the printers registry
# :Created:   sab 05 ago 2017 10:31:23 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017, 2018, 2019, 2021, 2022 Lele Gaifax
#

import warnings

import pytest

from pglast import ast, enums, prettify
from pglast.printers import IntEnumPrinter, NODE_PRINTERS, PrinterAlreadyPresentError
from pglast.printers import get_printer_for_node, node_printer


def test_registry():
    with pytest.raises(ValueError):
        get_printer_for_node(None)

    with pytest.raises(ValueError):
        @node_printer()
        def missing_node(node, output):
            pass

    with pytest.raises(ValueError):
        @node_printer(1)
        def invalid_node(node, output):
            pass

    with pytest.raises(ValueError):
        @node_printer(ast.RawStmt, ast.SelectStmt, ast.UpdateStmt)
        def too_many_nodes(node, output):
            pass

    with pytest.raises(ValueError):
        @node_printer('RawStmt')
        def invalid_node(node, output):
            pass

    with pytest.raises(ValueError):
        @node_printer((ast.RawStmt, 'foo'), ast.SelectStmt)
        def invalid_parents(node, output):
            pass

    raw_stmt = NODE_PRINTERS.pop(ast.RawStmt)
    try:
        @node_printer(ast.RawStmt)
        def raw(node, output):
            pass

        assert NODE_PRINTERS[ast.RawStmt] is raw

        with pytest.raises(PrinterAlreadyPresentError):
            @node_printer(ast.RawStmt)
            def other_raw(node, output):
                pass
    finally:
        NODE_PRINTERS[ast.RawStmt] = raw_stmt


def test_prettify_safety_belt():
    raw_stmt_printer = NODE_PRINTERS.pop(ast.RawStmt, None)
    try:
        @node_printer(ast.RawStmt)
        def raw_stmt_1(node, output):
            output.write('Yeah')

        output = prettify('select 42')
        assert output == 'Yeah'

        with warnings.catch_warnings(record=True) as w:
            output = prettify('select 42', safety_belt=True)
            assert output == 'select 42'
            assert 'Detected a bug' in str(w[0].message)

        @node_printer(ast.RawStmt, override=True)
        def raw_stmt_2(node, output):
            output.write('select 1')

        output = prettify('select 42')
        assert output == 'select 1'

        with warnings.catch_warnings(record=True) as w:
            output = prettify('select 42', safety_belt=True)
            assert output == 'select 42'
            assert 'Detected a non-cosmetic difference' in str(w[0].message)
    finally:
        if raw_stmt_printer is not None:
            NODE_PRINTERS[ast.RawStmt] = raw_stmt_printer
        else:
            NODE_PRINTERS.pop(ast.RawStmt, None)


def test_int_enum_printer():
    class LockWaitPrinter(IntEnumPrinter):
        enum = enums.LockWaitPolicy

        def LockWaitBlock(self, node, output):
            output.append('block')

    lwp = LockWaitPrinter()
    result = []
    lwp('LockWaitBlock', object(), result)
    assert result == ['block']

    with pytest.raises(NotImplementedError):
        lwp('LockWaitError', object(), result)

    with pytest.raises(ValueError):
        lwp('FooBar', object(), result)

    lwp(None, object(), result)
    assert result == ['block']*2


def test_not_int_enum_printer():
    class NotIntEnum(IntEnumPrinter):
        enum = enums.FunctionParameterMode

    with pytest.raises(ValueError):
        NotIntEnum()



# -*- coding: utf-8 -*-
# :Project:   pglast -- Tests on the visitors module
# :Created:   mar 11 mag 2021, 08:36:23
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2021, 2022 Lele Gaifax
#

import pytest

from pglast import ast, enums, parse_sql, visitors
from pglast.stream import RawStream


@pytest.mark.parametrize('stmt,rnames', (
    # DDL
    ('alter table foo add foreign key (x) references bar (id)', {'foo', 'bar'}),
    ('create table foo (int a, int b references f(id))', {'foo', 'f'}),
    ('create view foo.bar as select 1 from there', {'foo.bar', 'there'}),
    ('drop table foo.bar, bar.foo', {'foo.bar', 'bar.foo'}),
    ('drop view foo.bar, bar.foo', {'foo.bar', 'bar.foo'}),
    ('drop view "my.schema".bar, bar."my.table", "foo.bar"',
     {'"my.schema".bar', 'bar."my.table"', '"foo.bar"'}),

    # DML
    ('insert into foo values (1)', {'foo'}),
    ('insert into foo select * from bar', {'foo', 'bar'}),
    ('select 1', set()),
    ('select 1 from schemata.relation', {'schemata.relation'}),
    ('select a from b.c.d', {'b.c.d'}),
    ('select * from "my.schema"."my.table"', {'"my.schema"."my.table"'}),
    ('update foo set a=1', {'foo'}),

    # CTE
    ('with q1(x,y) as (select 1,2) select * from q1, q2', {'q2'}),
    ('with my_ref as (select * from my_ref where a=1) select * from my_ref',
     {'my_ref'}),
    ('with cte1 as (select 1), cte2 as (select * from cte1) select * from cte2',
     set()),
    ('''
     with recursive t(n) as (values (1) union all select n+1 from t where n < 100)
       select sum(n) from t
     ''', set()),
    ('''
     with cte1 as (select 1)
       select * from (with cte2 as (select * from cte1)
                        select * from cte2) as a
     ''', set()),
    ('''
     with to_archive as (delete from products where date < '2010-11-01' returning *)
       insert into products_log select * from to_archive
     ''', {'products', 'products_log'}),
    ('with "foo.bar" as (select * from tab) select * from "foo.bar"', {'tab'}),
    ('select (with my_ref as (select 1) select 1) from my_ref', {'my_ref'}),
    ('''
     with cte1 as (select 1)
        , cte2 as (with cte1 as (select * from cte1) select * from cte1)
       select * from cte2
     ''', set()),
    ('''
     with recursive t1 as (
       insert into yy select * from t2 returning *
     ), t2 as (
       insert into y select * from y returning *
     )
     select 1;
     ''', {'y', 'yy'}),
    ('''
     with recursive t(a) as (
       select 11
       union all
       select a+1 from t where a < 50
     )
     delete from y using t where t.a = y.a returning y.a
     ''', {'y'}),
    ('''
     with t as (
         delete from y
         where a <= 10
         returning *
     )
     select * from t
     ''', {'y'}),
    ('''
     select count(*) from (
       with q1(x) as (select random() from generate_series(1, 5))
         select * from q1
       union
         select * from q1
     ) ss
     ''', set())
))
def test_referenced_tables(stmt, rnames):
    assert visitors.referenced_relations(stmt) == rnames


def test_visiting_path():
    from types import SimpleNamespace as SN

    root = SN(list=[SN(a='a'), SN(b='b')])
    proot = visitors.Ancestor()
    assert proot @ root is root
    assert repr(proot) == 'ROOT'

    plist = proot / (root.list, 'list')
    assert plist @ root is root.list
    assert plist[0] is root.list
    assert plist[1] is None
    assert repr(plist) == 'ROOT → list'

    pa = plist / (root.list[0], 0)
    assert pa @ root is root.list[0]
    assert pa[0] is root.list[0]
    assert pa[1] is root.list
    assert pa[2] is None
    assert repr(pa) == 'ROOT → list → 0'

    pb = plist / (root.list[1], 1)
    assert pb @ root is root.list[1]
    assert pb[0] is root.list[1]
    assert pb[1] is root.list
    assert pb[2] is None
    assert repr(pb) == 'ROOT → list → 1'


def test_invalid_argument():
    with pytest.raises(ValueError):
        visitors.Visitor()('foo')


def test_empty_visitor():
    v = visitors.Visitor()
    assert v(()) is None


def test_ancestors():
    class Checker(visitors.Visitor):
        def visit_RawStmt(self, ancestors, node):
            pass

        def visit(self, ancestors, node):
            assert ancestors@self.root is node
            assert ast.RawStmt in ancestors
            assert ast.InsertStmt not in ancestors

    checker = Checker()

    raw = parse_sql('select 1')
    checker(raw)

    raw = parse_sql('SELECT * FROM ROWS FROM(generate_series(10,11), get_users())')
    checker(raw)


def test_closest_node():
    class Checker(visitors.Visitor):
        def visit(self, ancestors, node):
            if isinstance(node, ast.FuncCall):
                assert isinstance(ancestors[0], tuple)
                assert isinstance(abs(ancestors).node, ast.RangeFunction)

    raw = parse_sql('SELECT * FROM ROWS FROM(generate_series(10,11), get_users())')
    Checker()(raw)


class CountAllNodes(visitors.Visitor):
    def __call__(self, node):
        self.count = 0
        super().__call__(node)
        return self.count

    def visit(self, ancestors, node):
        self.count += 1


def test_count_all_nodes():
    counter = CountAllNodes()

    raw = parse_sql('select 1')
    assert counter(raw) == 5
    assert counter(raw[0].stmt) == 4

    raw = parse_sql('SELECT * FROM ROWS FROM(generate_series(10,11), get_users())')
    assert counter(raw) == 14


def test_skip_action():
    sql = 'select 1 where a = 1'

    all_nodes_count = CountAllNodes()(parse_sql(sql))
    assert all_nodes_count == 11

    class CountNodesSkippingExprs(visitors.Visitor):
        def __call__(self, node):
            self.count = 0
            self.exprs = 0
            super().__call__(node)
            return self.count, self.exprs

        def visit(self, ancestors, node):
            if node.__class__.__name__ == 'A_Expr':
                self.exprs += CountAllNodes()(node)
                return visitors.Skip
            self.count += 1

    counter = CountNodesSkippingExprs()

    raw = parse_sql(sql)
    nodes, exprs = counter(raw)
    assert nodes == all_nodes_count - exprs

    class CountNodesSkippingTargets(visitors.Visitor):
        def __call__(self, node):
            self.count = 0
            self.targets = 0
            super().__call__(node)
            return self.count, self.targets

        def visit(self, ancestors, node):
            if node.__class__.__name__ == 'ResTarget':
                self.targets += CountAllNodes()(node)
                return visitors.Skip
            self.count += 1

    counter = CountNodesSkippingTargets()

    raw = parse_sql('select 1 where a = 1')
    nodes, targets = counter(raw)
    assert nodes == all_nodes_count - targets


def test_delete_action():
    class DropNullConstraint(visitors.Visitor):
        def visit_Constraint(self, ancestors, node):
            if node.contype == enums.ConstrType.CONSTR_NULL:
                return visitors.Delete

    raw = parse_sql('create table foo (a integer null, b integer not null)')
    DropNullConstraint()(raw)
    assert RawStream()(raw) == 'CREATE TABLE foo (a integer, b integer NOT NULL)'

    raw = parse_sql('create table foo (a integer null check (a <> 0))')
    DropNullConstraint()(raw)
    assert RawStream()(raw) == 'CREATE TABLE foo (a integer CHECK (a <> 0))'

    class DeleteOddsInList(visitors.Visitor):
        def visit_A_Const(self, ancestors, node):
            if isinstance(node.val, ast.Integer):
                if node.val.ival % 2:
                    return visitors.Delete

    raw = parse_sql('INSERT INTO foo VALUES (1, 2, 3, 42, 43)')
    DeleteOddsInList()(raw)
    assert RawStream()(raw) == 'INSERT INTO foo VALUES (2, 42)'

    raw = parse_sql('INSERT INTO foo VALUES ((1, 2, 3, 42, 43),'
                    ' (2, 1, 4, 3, 5))')
    DeleteOddsInList()(raw)
    assert RawStream()(raw) == 'INSERT INTO foo VALUES ((2, 42), (2, 4))'

    raw = parse_sql('select true from foo where a in (1, 2, 3)')
    DeleteOddsInList()(raw)
    assert RawStream()(raw) == 'SELECT TRUE FROM foo WHERE a IN (2)'


def test_alter_node():
    class AddNullConstraint(visitors.Visitor):
        def visit_ColumnDef(self, ancestors, node):
            if node.colname == 'a':
                node.constraints = (
                    {'@': 'Constraint',
                     'contype': {'#': 'ConstrType', 'name': 'CONSTR_NOTNULL'}},)

        def visit_Constraint(self, ancestors, node):
            if ancestors[1].colname == 'b':
                return ast.Constraint(
                    contype=enums.ConstrType.CONSTR_CHECK,
                    raw_expr=ast.A_Expr(kind=enums.A_Expr_Kind.AEXPR_OP,
                                        lexpr=ast.ColumnRef(fields=(ast.String('b'),)),
                                        name=(ast.String('>'),),
                                        rexpr=ast.A_Const(val=ast.Integer(0))))

    raw = parse_sql('create table foo (a integer null, b integer check (b <> 0))')
    AddNullConstraint()(raw)
    assert RawStream()(raw) == 'CREATE TABLE foo (a integer NOT NULL, b integer CHECK (b > 0))'

    class DoubleAllIntegers(visitors.Visitor):
        def visit_Integer(self, ancestors, node):
            return ast.Integer(node.ival * 2)

    raw = parse_sql('select 21')
    DoubleAllIntegers()(raw)
    assert RawStream()(raw) == 'SELECT 42'

    class ReplaceConstantInList(visitors.Visitor):
        def visit_A_Const(self, ancestors, node):
            return ast.A_Const(val=ast.Integer(0))

    raw = parse_sql('INSERT INTO foo VALUES (42)')
    ReplaceConstantInList()(raw)
    assert RawStream()(raw) == 'INSERT INTO foo VALUES (0)'


def test_replace_root_node():
    class AndNowForSomethingCompletelyDifferent(visitors.Visitor):
        def visit_RawStmt(self, ancestors, node):
            return ast.RawStmt(stmt=ast.VariableShowStmt(name='all'))

    raw = parse_sql('select 1')
    new_root = AndNowForSomethingCompletelyDifferent()(raw)
    assert RawStream()(new_root) == 'SHOW ALL'

    raw = parse_sql('select 1')
    new_root = AndNowForSomethingCompletelyDifferent()(raw[0])
    assert RawStream()(new_root) == 'SHOW ALL'



# -*- coding: utf-8 -*-
# :Project:   pglast -- Extract enums from PostgreSQL headers
# :Created:   gio 03 ago 2017 14:54:39 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017, 2018, 2019, 2020, 2021, 2022 Lele Gaifax
#

from datetime import date
from os.path import basename, splitext
from re import match
import subprocess

from pycparser import c_ast, c_parser

try:
    from pglast.parser import LONG_MAX
except ModuleNotFoundError:
    # bootstrap
    from sys import maxsize as LONG_MAX


PY_HEADER = """\
# -*- coding: utf-8 -*-
# :Project:   pglast -- DO NOT EDIT: automatically extracted from %%s @ %%s
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017-%d Lele Gaifax
#

from enum import Enum, IntEnum, IntFlag, auto

try:
    from enum import StrEnum
except ImportError:  # pragma: no cover
    # Python < 3.10
    class StrEnum(str, Enum):
        pass

""" % date.today().year

RST_HEADER = """\
.. -*- coding: utf-8 -*-
.. :Project:   pglast -- DO NOT EDIT: generated automatically
.. :Author:    Lele Gaifax <lele@metapensiero.it>
.. :License:   GNU General Public License version 3 or later
.. :Copyright: © 2017-%d Lele Gaifax
..

========================================================%%(extra_decoration)s
 :mod:`pglast.enums.%%(mod_name)s` --- Constants extracted from `%%(header_fname)s`__
========================================================%%(extra_decoration)s

__ %%(header_url)s

.. module:: pglast.enums.%%(mod_name)s
   :synopsis: Constants extracted from %%(header_fname)s
""" % date.today().year


def get_libpg_query_info():
    "Return a tuple with (version, baseurl) of the libpg_query library."

    version = subprocess.check_output(['git', 'describe', '--all', '--long'],
                                      cwd='libpg_query')
    version = version.decode('utf-8').strip().split('/')[-1]
    remote = subprocess.check_output(['git', 'remote', 'get-url', 'origin'],
                                     cwd='libpg_query')
    remote = remote.decode('utf-8')
    baseurl = '%s/blob/%s/' % (remote[:-5], version[-7:])
    return version, baseurl


def preprocess(fname, cpp_args=[]):
    "Preprocess the given header and return the result."

    result = subprocess.check_output(['cpp', '-E', '-include', 'c.h', *cpp_args, fname])

    return result.decode('utf-8')


def extract_toc(header):
    "Extract the enums and defines with their position in the header."

    toc = {}

    with open(header, encoding='utf-8') as f:
        content = f.read()

    in_typedef_enum = 0

    for lineno, line in enumerate(content.splitlines(), 1):
        if line.startswith('typedef enum '):
            m = match(r'typedef enum\s+([\w_]+)', line)
            if m is not None:
                toc[m.group(1)] = lineno
        elif line.startswith('typedef enum'):
            in_typedef_enum = lineno
        elif in_typedef_enum and line.startswith('}'):
            m = match(r'}\s+([\w_]+)\s*;', line)
            if m is not None:
                toc[m.group(1)] = in_typedef_enum
                in_typedef_enum = 0
        elif line.startswith('#define'):
            m = match(r'#define\s+([a-zA-Z_]+)', line)
            if m is not None:
                toc[m.group(1)] = lineno

    return toc


def extract_enums(toc, source):
    "Yield all enum definitions belonging to the given header."

    typedefs = []
    in_typedef = False
    typedef = []

    for line in source.splitlines():
        if line and not line.startswith('#'):
            if in_typedef:
                typedef.append(line)
                if line.startswith('}'):
                    in_typedef = False
                    typedefs.append(typedef)
                    typedef = []
            elif line.startswith('typedef enum'):
                in_typedef = True
                typedef.append(line)

    parser = c_parser.CParser()
    for typedef in typedefs:
        td = parser.parse(''.join(typedef))
        if td.ext[0].name in toc:
            yield td


def extract_defines(source):
    "Yield all #defined constants in the given header."

    for line in source.splitlines():
        if line and line.startswith('#define'):
            m = match(r"#define\s+([a-zA-Z_]+)\s+\(?(\d+\s*<<\s*\d+|(0x)?\d+|'[a-zA-Z]')\)?",
                      line)
            if m is not None:
                yield m.group(1), m.group(2)
            else:
                m = match(r"#define\s+([a-zA-Z_]+)\s+LONG_MAX", line)
                if m is not None:
                    yield m.group(1), LONG_MAX


def emit_constant(value):
    return value.value


def emit_binary_op(value):
    assert isinstance(value.left, c_ast.Constant)
    assert isinstance(value.right, c_ast.Constant)
    return '%s %s %s' % (emit_constant(value.left),
                         value.op,
                         emit_constant(value.right))


def emit_unary_op(value):
    return '%s%s' % (value.op, emit_constant(value.expr))


def int_enum_value_factory(index, enumerator):
    if enumerator.value is None:
        return '0' if index == 0 else 'auto()'

    if isinstance(enumerator.value, c_ast.BinaryOp):
        return emit_binary_op(enumerator.value)
    elif isinstance(enumerator.value, c_ast.Constant):
        return emit_constant(enumerator.value)
    elif isinstance(enumerator.value, c_ast.UnaryOp):
        return emit_unary_op(enumerator.value)
    elif enumerator.value.name == 'PG_INT32_MAX':
        return '0x7FFFFFFF'

    assert enumerator.value.type == 'int'
    return enumerator.value.value


def char_enum_value_factory(index, enumerator):
    assert enumerator.value.type == 'char'
    return enumerator.value.value


def determine_enum_type_and_value(enum):
    type = 'IntEnum'
    value = int_enum_value_factory

    for item in enum.values.enumerators:
        if item.value:
            if isinstance(item.value, c_ast.Constant) and item.value.type == 'char':
                type = 'StrEnum'
                value = char_enum_value_factory
                break
            elif isinstance(item.value, c_ast.BinaryOp) and item.value.op == '<<':
                type = 'IntFlag'
                break

    return type, value


def write_enum(name, enum, output):
    enum_type, value_factory = determine_enum_type_and_value(enum)
    output.write('\n')
    output.write('class %s(%s):\n' % (name, enum_type))
    for index, item in enumerate(enum.values.enumerators):
        output.write('    %s = %s\n' % (item.name, value_factory(index, item)))


def write_enum_doc(name, enum, output, toc, url, mod_name):
    output.write('\n\n.. class:: pglast.enums.%s.%s\n' % (mod_name, name))
    if name in toc:
        output.write('\n   Corresponds to the `%s enum <%s#L%d>`__.\n' %
                     (name, url, toc[name]))
    for index, item in enumerate(enum.values.enumerators):
        output.write('\n   .. data:: %s\n' % item.name)


def workhorse(args):
    libpg_query_version, libpg_query_baseurl = get_libpg_query_info()
    header_url = libpg_query_baseurl + args.header[12:]
    toc = extract_toc(args.header)
    preprocessed = preprocess(args.header, ['-I%s' % idir for idir in args.include_directory])
    with open(args.output, 'w', encoding='utf-8') as output, \
         open(args.rstdoc, 'w', encoding='utf-8') as rstdoc:
        header_fname = basename(args.header)
        mod_name = splitext(header_fname)[0]
        output.write(PY_HEADER % (header_fname, libpg_query_version))
        rstdoc.write(RST_HEADER % dict(
            mod_name=mod_name, header_fname=header_fname,
            extra_decoration='='*(len(mod_name) + len(header_fname)),
            header_url=header_url))

        for node in sorted(extract_enums(toc, preprocessed),
                           key=lambda x: x.ext[0].name):
            enum = node.ext[0].type.type
            write_enum(enum.name or node.ext[0].name, enum, output)
            write_enum_doc(enum.name or node.ext[0].name, enum, rstdoc, toc, header_url,
                           mod_name)

        separator_emitted = False
        with open(args.header, encoding='utf-8') as header:
            for constant, value in extract_defines(header.read()):
                if not separator_emitted:
                    output.write('\n\n')
                    output.write('# #define-ed constants\n')
                    rstdoc.write('\n')
                    separator_emitted = True
                output.write('\n%s = %s\n' % (constant, value))
                rstdoc.write('\n.. data:: %s\n' % constant)
                if constant in toc:
                    rstdoc.write('\n   See `here for details <%s#L%d>`__.\n'
                                 % (header_url, toc[constant]))


def main():
    from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

    parser = ArgumentParser(description="PG enum extractor",
                            formatter_class=ArgumentDefaultsHelpFormatter)

    parser.add_argument('-I', '--include-directory', action='append', metavar='DIR',
                        help="add DIR to the list of include directories")
    parser.add_argument('header',
                        help="source header to be processed")
    parser.add_argument('output',
                        help="Python source to be created")
    parser.add_argument('rstdoc',
                        help="reST documentation to be created")

    args = parser.parse_args()

    workhorse(args)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# :Project:   pglast -- Extract keywords from PostgreSQL header
# :Created:   dom 06 ago 2017 23:34:53 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017, 2018, 2019, 2021, 2022 Lele Gaifax
#

from collections import defaultdict
from os.path import basename
from pprint import pformat
from re import match


HEADER = """\
# -*- coding: utf-8 -*-
# :Project:   pglast -- DO NOT EDIT: automatically extracted from %s @ %s
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017 Lele Gaifax
#
"""


def get_target_pg_version():
    "Return the target PG version"

    with open('libpg_query/Makefile') as mf:
        for line in mf:
            if line.startswith('PG_VERSION ='):
                pg_version = line.split('=')[1].strip()
                break
        else:
            raise RuntimeError('Could not determine target PG version, "PG_VERSION" not found'
                               ' in libpg_query/Makefile!')

    return pg_version


def extract_keywords(source):
    for line in source.splitlines():
        if line.startswith('PG_KEYWORD'):
            m = match(r'PG_KEYWORD\("([^"]+)",[^,]+,\s*([\w_]+),[^)]+\)', line.strip())
            if m:
                yield m.group(1), m.group(2)


def workhorse(args):
    with open(args.header, encoding='utf-8') as f:
        source = f.read()

    bytype = defaultdict(set)
    for keyword, type in extract_keywords(source):
        bytype[type].add(keyword)

    with open(args.output, 'w', encoding='utf-8') as output:
        output.write(HEADER % (basename(args.header), get_target_pg_version()))
        for type in sorted(bytype):
            output.write('\n')
            output.write(type + 'S')
            output.write(' = {')
            keywords = pformat(bytype[type], compact=True, indent=len(type)+5, width=95)
            output.write(keywords[1:].lstrip())
            output.write('\n')


def main():
    from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

    parser = ArgumentParser(description="PG keyword extractor",
                            formatter_class=ArgumentDefaultsHelpFormatter)

    parser.add_argument('header',
                        help="source header to be processed")
    parser.add_argument('output',
                        help="Python source to be created")

    args = parser.parse_args()

    workhorse(args)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# :Project:   pglast -- Simple frontend to the pretty reformatter
# :Created:   dom 06 ago 2017 23:09:23 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017, 2018, 2019, 2021 Lele Gaifax
#

import argparse
import json
import pprint
import sys

from pglast import Error, parse_plpgsql, parse_sql, prettify


def workhorse(args):
    if args.statement:
        statement = args.statement
    else:
        input = args.infile or sys.stdin
        with input:
            statement = input.read()

    if args.parse_tree or args.plpgsql:
        tree = parse_plpgsql(statement) if args.plpgsql else parse_sql(statement)
        output = args.outfile or sys.stdout
        with output:
            if args.plpgsql:
                json.dump(tree, output, sort_keys=True, indent=2)
            else:
                pprint.pprint([stmt(skip_none=True) for stmt in tree], output)
            output.write('\n')
    else:
        try:
            prettified = prettify(
                statement,
                preserve_comments=args.preserve_comments,
                compact_lists_margin=args.compact_lists_margin,
                split_string_literals_threshold=args.split_string_literals,
                special_functions=args.special_functions,
                comma_at_eoln=args.comma_at_eoln,
                remove_pg_catalog_from_functions=args.remove_pg_catalog_from_functions,
                semicolon_after_last_statement=args.semicolon_after_last_statement)
        except Error as e:
            print()
            raise SystemExit(e)

        output = args.outfile or sys.stdout
        with output:
            output.write(prettified)
            output.write('\n')


def main(options=None):
    from argparse import ArgumentParser
    from pkg_resources import get_distribution
    from .parser import get_postgresql_version

    version = '%s, with PostgreSQL %s parser' % (
        get_distribution('pglast').version,
        '.'.join(str(p) for p in get_postgresql_version()))

    parser = ArgumentParser(description="PostgreSQL language prettifier")

    parser.add_argument('-v', '--version', action='version', version='%(prog)s ' + version,)
    parser.add_argument('-p', '--plpgsql', action='store_true', default=False,
                        help='use the plpgsql parser (and print just the resulting tree)')
    parser.add_argument('-t', '--parse-tree', action='store_true', default=False,
                        help='show just the parse tree of the statement')
    parser.add_argument('-m', '--compact-lists-margin', type=int, default=0,
                        help='use compact form for lists not exceeding the given margin')
    parser.add_argument('-s', '--split-string-literals', type=int, default=0,
                        help='split string literals longer than given value')
    parser.add_argument('-f', '--special-functions', action='store_true', default=False,
                        help='activate special functions handling')
    parser.add_argument('-F', '--remove-pg_catalog-from-functions', action='store_true',
                        default=False,
                        help='omit explicit "pg_catalog" schema from function names,'
                        ' when possible')
    parser.add_argument('-c', '--comma-at-eoln', action='store_true', default=False,
                        help='use alternative style to print lists, putting the comma right'
                        ' after each item')
    parser.add_argument('-e', '--semicolon-after-last-statement', action='store_true',
                        default=False, help='end the last statement with a semicolon')
    parser.add_argument('-C', '--preserve-comments', action='store_true',
                        default=False, help="preserve comments in the statement")
    parser.add_argument('-S', '--statement',
                        help='the SQL statement')
    parser.add_argument('infile', nargs='?', type=argparse.FileType(),
                        help='a file containing the SQL statement to be pretty-printed,'
                        ' by default stdin, when not specified with --statement option')
    parser.add_argument('outfile', nargs='?', type=argparse.FileType('w'),
                        help='where the result will be written, by default stdout')

    args = parser.parse_args(options if options is not None else sys.argv[1:])

    workhorse(args)


if __name__ == '__main__':  # pragma: no cover
    main()



# -*- coding: utf-8 -*-
# :Project:   pglast -- Tests on the stream.py module
# :Created:   sab 05 ago 2017 10:31:23 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017, 2018, 2019, 2021, 2022 Lele Gaifax
#

import pytest

from pglast import ast, parse_sql
from pglast.printers import NODE_PRINTERS, PrinterAlreadyPresentError, SPECIAL_FUNCTIONS
from pglast.printers import node_printer, special_function
from pglast.stream import IndentedStream, OutputStream, RawStream


def test_output_stream():
    output = OutputStream()
    output.writes('SELECT *')
    output.writes(' FROM')
    output.writes('table ')
    output.writes('WHERE')
    output.write('id = 1')

    assert output.getvalue() == 'SELECT * FROM table WHERE id = 1'


def test_raw_stream_with_sql():
    raw_stmt_printer = NODE_PRINTERS.pop(ast.RawStmt, None)
    try:
        @node_printer(ast.RawStmt)
        def raw_stmt(node, output):
            output.write('Yeah')

        output = RawStream()
        result = output('SELECT 1; SELECT 2')
        assert result == 'Yeah; Yeah'
    finally:
        if raw_stmt_printer is not None:
            NODE_PRINTERS[ast.RawStmt] = raw_stmt_printer
        else:
            NODE_PRINTERS.pop(ast.RawStmt, None)


def test_raw_stream():
    raw_stmt_printer = NODE_PRINTERS.pop(ast.RawStmt, None)
    try:
        @node_printer(ast.RawStmt)
        def raw_stmt(node, output):
            output.write('Yeah')

        root = parse_sql('SELECT 1')
        output = RawStream()
        result = output(root)
        assert result == 'Yeah'
    finally:
        if raw_stmt_printer is not None:
            NODE_PRINTERS[ast.RawStmt] = raw_stmt_printer
        else:
            NODE_PRINTERS.pop(ast.RawStmt, None)


def test_raw_stream_invalid_call():
    with pytest.raises(ValueError):
        RawStream()(1)


def test_indented_stream_with_sql():
    raw_stmt_printer = NODE_PRINTERS.pop(ast.RawStmt, None)
    try:
        @node_printer(ast.RawStmt)
        def raw_stmt(node, output):
            output.write('Yeah')

        output = IndentedStream()
        result = output('SELECT 1; SELECT 2')
        assert result == 'Yeah;\n\nYeah'

        output = IndentedStream(separate_statements=False)
        result = output('SELECT 1; SELECT 2')
        assert result == 'Yeah;\nYeah'
    finally:
        if raw_stmt_printer is not None:
            NODE_PRINTERS[ast.RawStmt] = raw_stmt_printer
        else:
            NODE_PRINTERS.pop(ast.RawStmt, None)


def test_separate_statements():
    """Separate statements by ``separate_statements`` (int) newlines."""
    raw_stmt_printer = NODE_PRINTERS.pop(ast.RawStmt, None)
    try:
        @node_printer(ast.RawStmt)
        def raw_stmt(node, output):
            output.write('Yeah')

        output = IndentedStream(separate_statements=2)
        result = output('SELECT 1; SELECT 2')
        assert result == 'Yeah;\n\n\nYeah'
    finally:
        if raw_stmt_printer is not None:
            NODE_PRINTERS[ast.RawStmt] = raw_stmt_printer
        else:
            NODE_PRINTERS.pop(ast.RawStmt, None)


def test_special_function():
    output = RawStream(special_functions=True)

    assert output.get_printer_for_function('foo.test_function') is None

    try:
        @special_function('foo.test_function')
        def test(node, output):
            pass

        assert output.get_printer_for_function('foo.test_function') is test

        with pytest.raises(PrinterAlreadyPresentError):
            @special_function('foo.test_function')
            def test1(node, output):
                pass

        @special_function('foo.test_function', override=True)
        def test_function(node, output):
            output.print_list(node.args, '-')

        result = output('SELECT foo.test_function(x, "Y") FROM sometable')
        assert result == 'SELECT x - "Y" FROM sometable'
    finally:
        SPECIAL_FUNCTIONS.pop('foo.test_function')



# -*- coding: utf-8 -*-
# :Project:   pglast -- Top level error exception class
# :Created:   mer 02 ago 2017 15:35:39 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017, 2018 Lele Gaifax
#

class Error(Exception):
    "Top level error exception."



# -*- coding: utf-8 -*-
# :Project:   pglast -- Visitors base machinery
# :Created:   dom 9 mag 2021, 16:15:05
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2021, 2022 Lele Gaifax
#

from collections import deque
from inspect import getmembers, ismethod

from . import ast
from .stream import maybe_double_quote_name


class ActionMeta(type):
    "Metaclass used to implement action singleton."

    def __repr__(cls):
        return cls.__name__


class Action(metaclass=ActionMeta):
    "Abstract action singleton."

    def __new__(cls):  # pragma: no cover
        return cls


class Add(Action):
    "Marker used to tell the iterator to insert nodes in the current sequence."


class Continue(Action):
    "Marker used to tell the iterator to keep going."


class Delete(Action):
    "Marker used to tell the iterator to delete current node."


class Skip(Action):
    "Marker used to tell the iterator to not descend into the current node."


class Ancestor:
    """Simple object to keep track of the node's ancestors while it's being visited.

    :param Ancestor parent: the parent of the new instance
    :param node: the tracked object
    :param member: either the name of the attribute or the index in the sequence
                   that points to the tracked object in the parent container

    An instance of this class represent a particular ancestor in the hierarchy chain: it
    carries a reference that points to the higher item in the chain, the associated
    :class:`.ast.Node` instance and a *member*, either the attribute name or sequential index
    in the parent node: the latter happens when the parent node is actually a tuple, not an
    ``Node`` instance.

    Accessing an instance with a positive index returns the nth node up in the hierarchy.

    When applied (using the ``@`` operator) to an :class:`.ast.Node` instance will traverse
    that node returning the leaf one corresponding to the whole chain.

    Example:

    .. testsetup:: *

        from pglast import parse_sql
        from pglast.visitors import Ancestor

    .. doctest::

        >>> tree = parse_sql('select 1')
        >>> root = Ancestor()
        >>> root
        ROOT
        >>> root@tree is tree
        True
        >>> root[0] is None
        True
        >>> select_stmt_path = root / (tree, 0) / (tree[0], 'stmt')
        >>> select_stmt_path
        ROOT → 0 → stmt
        >>> select_stmt_path@tree is tree[0].stmt
        True
        >>> select_stmt_path[0] is tree[0]
        True
        >>> columns_path = (select_stmt_path
        ...                 / (tree[0].stmt, 'targetList'))
        >>> first_col_path = (columns_path
        ...                   / (tree[0].stmt.targetList[0], 0))
        >>> first_col_path
        ROOT → 0 → stmt → targetList → 0
        >>> first_col_path[0]
        <ResTarget val=<A_Const isnull=False val=<Integer ival=1>>>
        >>> first_col_path[1] is columns_path[0]
        True

    As said, the `node` is not always a :class:`.ast.Node`, but may be a tuple, possibly
    containing *subtuples*, for example the ``functions`` slot of :class:`~.ast.RangeFunction`
    that is a tuple of tuples, each containing a :class:`~.ast.FuncCall` and possibly other
    values:

    .. testsetup:: *

        from pglast.visitors import Visitor

    .. doctest::

        >>> tree = parse_sql('SELECT * FROM ROWS'
        ...                  ' FROM (generate_series(10,11),'
        ...                  '       get_users())')
        >>> class VerboseVisitor(Visitor):
        ...   all_ancestors = []
        ...   def visit(self, ancestors, node):
        ...     print(f'{len(self.all_ancestors):2d}.',
        ...           "node:", type(node).__name__,
        ...           "ancestor:", type(ancestors.node).__name__)
        ...     self.all_ancestors.append(ancestors)
        >>> VerboseVisitor()(tree)
         0. node: RawStmt ancestor: tuple
         1. node: SelectStmt ancestor: RawStmt
         2. node: ResTarget ancestor: tuple
         3. node: RangeFunction ancestor: tuple
         4. node: ColumnRef ancestor: ResTarget
         5. node: A_Star ancestor: tuple
         6. node: FuncCall ancestor: tuple
         7. node: FuncCall ancestor: tuple
         8. node: String ancestor: tuple
         9. node: A_Const ancestor: tuple
        10. node: A_Const ancestor: tuple
        11. node: String ancestor: tuple
        12. node: Integer ancestor: A_Const
        13. node: Integer ancestor: A_Const
        ...

    To find the closest ancestor that is actually pointing to an AST node you may use the
    ``abs()`` function:

    .. doctest::

        >>> range_function = tree[0].stmt.fromClause[0]
        >>> gen_series_funccall = range_function.functions[0][0]
        >>> gen_series_funccall
        <FuncCall funcname=(<String sval='generate_series'>,) ...>
        >>> generate_series_ancestry = VerboseVisitor.all_ancestors[6]
        >>> generate_series_ancestry@tree is gen_series_funccall
        True
        >>> abs(generate_series_ancestry).node is range_function
        True

    As an aid to visitors that apply changes to the AST, there are two methods, :meth:`update`
    and :meth:`apply`, that takes care of the different cases, when `node` is an AST instance
    or instead it's a tuple (or subtuple); the former does not directly change the AST tree,
    but postpones that when the latter is called.
    """

    __slots__ = ('parent', 'node', 'member', 'pending_update')

    def __init__(self, parent=None, node=None, member=None):
        self.parent = parent
        self.node = node
        self.member = member
        self.pending_update = None

    def _iter_members(self):
        "Return the sequence of `member` of the whole hierarchy."

        ancestors = []

        path = self
        while True:
            ancestors.append(path.member)
            if path.parent is None:
                return reversed(ancestors)
            path = path.parent

    def __repr__(self):
        return ' → '.join('ROOT' if m is None else str(m) for m in self._iter_members())

    def __getitem__(self, n):
        path = self
        while n > 0:
            path = path.parent
            n -= 1
        return path.node

    def __abs__(self):
        path = self
        while path:
            if isinstance(path.node, ast.Node):
                return path
            path = path.parent

    def find_nearest(self, cls):
        "Find the nearest ancestor with a node of the given `cls`."

        path = self
        while True:
            if isinstance(path.node, cls):
                return path
            path = path.parent
            if path is None:
                break

    def __contains__(self, cls):
        "Tell whether there is a node of type `cls` in the anchestry."

        return self.find_nearest(cls) is not None

    def __truediv__(self, node_and_member):
        "Create a new instance pointing to the given child node."

        return Ancestor(self, *node_and_member)

    def __matmul__(self, root):
        "Resolve the ancestry chain against the given `root` node, returning the leaf node."

        node = root
        for member in self._iter_members():
            if member is not None:
                if isinstance(member, int):
                    node = node[member]
                else:
                    node = getattr(node, member)
        return node

    def update(self, new_value):
        "Set `new_value` as a pending change to the tracked node."

        if isinstance(self.member, int):
            # We are pointing to one particular item in a tuple, so we need to build a new one,
            # replacing it with the new value and then update the parent node accordingly:
            # since there may be further changes, temporarily use a list instead of a tuple, at
            # apply() time it will be coerced back to a tuple. Note that there are "list of
            # lists" cases, that are handled at apply time.
            if self.parent.pending_update is None:
                self.parent.pending_update = list(self.node)
            self.parent.pending_update[self.member] = new_value
            return self.parent
        else:
            self.pending_update = new_value
            return self

    def apply(self):
        "Apply the pending change, if any, to the actual node."

        if self.pending_update is not None:
            if isinstance(self.pending_update, list):
                value = tuple(filter(lambda item: item is not Delete, self.pending_update))
            else:
                value = self.pending_update
            if isinstance(self.member, int):
                pvalue = list(self.node)
                pvalue[self.member] = value or None
                self.parent.update(pvalue)
                self.parent.apply()
            else:
                if self.member is None:
                    self.node = value or None
                else:
                    setattr(self.node, self.member, value or None)
            self.pending_update = None


class Visitor:
    """Base class implementing the `visitor pattern`__.

    __ https://en.wikipedia.org/wiki/Visitor_pattern

    To use it, you shall write a subclass that implements a set of particular named methods,
    specifically ``visit_XYZ`` where ``XYZ`` is the name of a class name defined in the
    :mod:`pglast.ast` module.

    Instances of this class are *callables* and accept either a :class:`.ast.Node`
    instance or a sequence of instances, typically the result of :func:`parse_sql
    <pglast.parser.parse_sql>`. The argument will be *traversed* in a `breadth first`__ order
    and each :class:`Node <.ast.Node>` instance will be passed to the corresponding
    ``visit_XYZ`` method if it is implemented, falling back to the default ``visit`` method. If
    none of them are defined, the node will be ignored.

    The ``visit_XYZ`` methods receive two arguments: the *ancestry chain* of the node, an
    instance of :class:`Ancestor` and the :class:`Node <.ast.Node>` instance itself. The
    methods may return either ``None``, an *action* or a new node that will replace the
    original one.

    __ https://en.wikipedia.org/wiki/Breadth-first_search
    """

    def __call__(self, root):
        "Iteratively visit the `root` node calling related ``visit_XYZ`` methods."

        self.root = root

        by_ast_class = {}
        for name, method in getmembers(self, ismethod):
            if not name.startswith('visit_'):
                continue

            clsname = name[6:]
            cls = getattr(ast, clsname, None)
            if cls is not None:
                by_ast_class[cls] = method

        default_method = self.visit

        generator = self.iterate(root)
        try:
            ancestors, node = generator.send(None)
        except StopIteration:
            return

        while True:
            method = by_ast_class.get(node.__class__, default_method)
            if method is not None:
                result = method(ancestors, node)
            else:
                result = None
            try:
                ancestors, node = generator.send(Continue if result is None else result)
            except StopIteration:
                break

        return self.root

    def iterate(self, node):
        """Iterate thru `node`'s AST using a breadth-first traversing.

        :param node: either a :class:`.ast.Node` instance or a tuple of those

        This is a generator, that yields ``Node`` instances together with their *ancestors
        chain* as it finds them while traversing the tree.
        """

        pending_updates = []

        todo = deque()

        if isinstance(node, (tuple, ast.Node)):
            todo.append((Ancestor(), node))
        else:
            raise ValueError('Bad argument, expected a ast.Node instance or a tuple')

        while todo:
            ancestors, node = todo.popleft()

            # Here `node` may be either one AST node, a tuple of AST nodes (e.g.
            # SelectStmt.targetList), or even a tuple of tuples of AST nodes (e.g.
            # SelectStmt.valuesList). To simplify code, coerce it to a sequence.

            is_sequence = isinstance(node, tuple)
            if is_sequence:
                nodes = list(node)
            else:
                nodes = [node]

            index = 0
            while nodes:
                sub_node = nodes.pop(0)
                if is_sequence:
                    sub_ancestors = ancestors / (node, index)
                else:
                    sub_ancestors = ancestors
                if isinstance(sub_node, ast.Node):
                    action = yield sub_ancestors, sub_node
                    if action is Continue:
                        for member in sub_node:
                            value = getattr(sub_node, member)
                            if isinstance(value, (tuple, ast.Node)):
                                todo.append((sub_ancestors / (sub_node, member), value))
                    elif action is Skip:
                        pass
                    else:
                        pending_updates.append(sub_ancestors.update(action))
                elif isinstance(sub_node, tuple):
                    for sub_index, value in enumerate(sub_node):
                        if isinstance(value, (tuple, ast.Node)):
                            todo.append((sub_ancestors / (sub_node, sub_index), value))

                index += 1

        for pending_update in pending_updates:
            pending_update.apply()
            if pending_update.member is None:
                self.root = pending_update.node

    visit = None
    """
    The default *visit* method for any node without a specific one.
    When ``None``, nothing happens.
    """


class ReferencedRelations(Visitor):
    """Concrete implementation of the :func:`.referenced_relations` function.

    :param set cte_names: the set of surrounding CTE names
    :param WithClause skip_with_clause: skip this clause, when specified

    Calling an instance of this class will return a set of the names of the
    relations referenced by the given :class:`node <pglast.ast.Node>`.
    """

    def __init__(self, cte_names=None, skip_with_clause=None):
        super().__init__()
        self.cte_names = set() if cte_names is None else cte_names.copy()
        self.skip_with_clause = skip_with_clause
        self.r_names = set()

    def __call__(self, node):
        super().__call__(node)
        return self.r_names

    def visit_DropStmt(self, ancestors, node):
        from .enums import ObjectType

        if node.removeType in (ObjectType.OBJECT_TABLE, ObjectType.OBJECT_VIEW):
            for obj in node.objects:
                self.r_names.add('.'.join(maybe_double_quote_name(n.sval) for n in obj))

    def visit_SelectStmt(self, ancestors, node):
        if node.withClause and node.withClause is not self.skip_with_clause:
            # CTEs are tricky to get right, as issue #106 demonstrates!
            #
            # We must first collect the CTE names as well as the table referenced by their
            # queries, and then process the statement with that knowledge.
            #
            # In the normal case, each CTE must be processed in order, with the CTE names found
            # earlier; in the "WITH RECURSIVE" case all its CTEs must be considered valid at
            # the same time.
            with_clause = node.withClause
            if with_clause.recursive:
                self.cte_names.update(maybe_double_quote_name(cte.ctename)
                                      for cte in with_clause.ctes)
                self.r_names.update(ReferencedRelations(self.cte_names)(with_clause))
            else:
                for cte in with_clause.ctes:
                    cte_name = maybe_double_quote_name(cte.ctename)
                    self.r_names.update(ReferencedRelations(self.cte_names)(cte))
                    self.cte_names.add(cte_name)
            self.r_names.update(ReferencedRelations(self.cte_names, with_clause)(node))
            return Skip

    visit_UpdateStmt = visit_InsertStmt = visit_DeleteStmt = visit_SelectStmt

    def visit_WithClause(self, ancestors, node):
        if node is self.skip_with_clause:
            return Skip

    def visit_RangeVar(self, ancestors, node):
        "Collect relation names, taking into account defined CTE names"

        tname = maybe_double_quote_name(node.relname)

        if node.schemaname:
            tname = f'{maybe_double_quote_name(node.schemaname)}.{tname}'

        if node.catalogname:
            tname = f'{maybe_double_quote_name(node.catalogname)}.{tname}'

        if tname not in self.cte_names:
            self.r_names.add(tname)


def referenced_relations(stmt):
    """Return the set of relation names referenced in the given `stmt`.

    :param stmt: either a string containing the ``SQL`` statement or a :class:`.ast.Node`
                 instance
    :returns: a set of strings, the names of the relations

    Example:

    .. testsetup::

        from pglast.visitors import referenced_relations

    .. doctest::

        >>> referenced_relations('WITH q1(x,y) AS (SELECT 1,2)'
        ...                      ' SELECT * FROM q1, q2')
        {'q2'}
    """

    from .parser import parse_sql

    return ReferencedRelations()(parse_sql(stmt) if isinstance(stmt, str) else stmt)



# -*- coding: utf-8 -*-
# :Project:   pglast -- Printer functions for PG parse tree nodes
# :Created:   sab 05 ago 2017 16:33:14 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017, 2018, 2020, 2021, 2022 Lele Gaifax
#

from .. import ast
from ..error import Error


NODE_PRINTERS = {}
"Registry of specialized node printers, keyed by their class."


SPECIAL_FUNCTIONS = {}
"Registry of specialized function printers, keyed by their qualified name."


class PrinterAlreadyPresentError(Error):
    "Exception raised trying to register another function for a tag already present."


def get_printer_for_node(node):
    """Get specific printer implementation for given `node`.

    If there is a more specific printer for it, when it's inside a particular
    ancestor, return that instead.
    """

    node_class = type(node)
    if not issubclass(node_class, ast.Node):
        raise ValueError('Expected an ast.Node, not a %r' % node_class.__name__)
    parent = abs(node.ancestors)
    parent_node_class = None if parent is None else type(parent.node)
    printer = NODE_PRINTERS.get((parent_node_class, node_class))
    if printer is None:
        printer = NODE_PRINTERS.get(node_class)
        if printer is None:
            raise NotImplementedError("Printer for node %r is not implemented yet"
                                      % node_class.__name__)
    return printer


def node_printer(*nodes, override=False):
    r"""Decorator to register a specific printer implementation for a (set of) `nodes`.

    :param \*nodes: a list of one or two items
    :param bool override:
           when ``True`` the function will be registered even if already present in the
           :data:`NODE_PRINTERS` registry

    When `nodes` contains a single item then the decorated function is the *generic*
    one, and it will be registered in :data:`NODE_PRINTERS` with that key alone. Otherwise it
    must contain two elements: the first may be either a scalar value or a sequence of parent
    nodes, and the function will be registered under the key ``(parent, node)``.
    """

    n = len(nodes)
    if n == 1:
        parent_classes = (None,)
        node_class = nodes[0]
    elif n == 2:
        parent_classes, node_class = nodes
        if not isinstance(parent_classes, (list, tuple)):
            parent_classes = (parent_classes,)
        if not all(isinstance(c, type) and issubclass(c, ast.Node) for c in parent_classes):
            raise ValueError('Invalid nodes: expected a sequence of ast.Node classes as'
                             '  parents, got %r' % (parent_classes,))
    else:
        raise ValueError('Invalid nodes: must contain one or two items!')

    if not isinstance(node_class, type) or not issubclass(node_class, ast.Node):
        raise ValueError('Invalid nodes: expected an ast.Node class,'
                         ' got %r' % node_class)

    def decorator(impl):
        for parent_class in parent_classes:
            key = node_class if parent_class is None else (parent_class, node_class)
            if not override and key in NODE_PRINTERS:  # pragma: no cover
                raise PrinterAlreadyPresentError("A printer is already registered for tag %r"
                                                 % key)
            NODE_PRINTERS[key] = impl
        return impl

    return decorator


def special_function(name, override=False):
    """Decorator to declare a particular PostgreSQL function `name` as *special*, with a
    specific printer.

    :param str name: the qualified name of the PG function
    :param bool override: when ``True`` the function will be registered even if already
                          present in the :data:`SPECIAL_FUNCTIONS` registry
    """

    def decorator(impl):
        if not override and name in SPECIAL_FUNCTIONS:
            raise PrinterAlreadyPresentError("A printer is already registered for function %r"
                                             % name)
        SPECIAL_FUNCTIONS[name] = impl
        return impl
    return decorator


get_special_function = SPECIAL_FUNCTIONS.get


class IntEnumPrinter:
    "Helper class used to implement printers for specific enumerated values."

    enum = None
    "The enum class this is a printer for."

    def __init__(self):
        from enum import IntEnum

        if not issubclass(self.enum, IntEnum):
            raise ValueError(f'Given enum {self.enum!r} is not an IntEnum subclass')
        if len(set(m.value for m in self.enum)) != len(self.enum):  # pragma: no cover
            raise ValueError(f'Given enum {self.enum!r} contains aliased symbols')
        self.value_to_symbol = {m.value: m.name for m in self.enum}

    def __call__(self, value, node, output):
        if value is None:
            # Should never happen, but better safe than sorry
            for symbol, member in self.enum.__members__.items():
                if member.value == 0:
                    break
            else:  # pragma: no cover
                raise ValueError(f"Could not determine default value of class {self.enum!r}")
        elif isinstance(value, self.enum):
            symbol = self.value_to_symbol.get(value)
        elif isinstance(value, ast.Integer):
            symbol = self.value_to_symbol.get(value.ival)
        else:
            symbol = value
        if symbol is None:  # pragma: no cover
            raise ValueError(f"Invalid value {value!r}, not in class {self.enum!r}")
        method = getattr(self, symbol, None)
        if method is None:
            if symbol in self.enum.__members__:
                raise NotImplementedError(f"Printer for {symbol!r} of {self.enum!r} not"
                                          f" implemented yet")
            else:
                raise ValueError(f"Invalid symbol {symbol!r}, not in class {self.enum!r}")

        method(node, output)


def get_string_value(lst):
    "Helper function to get a literal string value, wrapped in a one-sized list."

    if len(lst) != 1 or not isinstance(lst[0], ast.String):  # pragma: no cover
        raise TypeError('%r does not contain a single String node' % lst)
    return lst[0].sval


from . import ddl, dml, sfuncs  # noqa: F401,E402
